# Copyright 2015
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_index_tenant_id

Revision ID: 4b6d8d5310b8
Revises: 4deef6d81931
Create Date: 2015-02-10 18:28:26.362881

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '4b6d8d5310b8'
down_revision = '4deef6d81931'

TABLES = ['lbaas_members', 'lbaas_healthmonitors', 'lbaas_pools',
          'lbaas_loadbalancers', 'lbaas_listeners', 'vips', 'members',
          'pools', 'healthmonitors']


def upgrade():
    for table in TABLES:
        op.create_index(op.f('ix_%s_tenant_id' % table),
                        table, ['tenant_id'], unique=False)



# Copyright 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
import webob

from neutron.db import servicetype_db as st_db
from neutron.quota import resource_registry
from neutron.tests import base as n_base
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit.extensions import base as ext_base
from neutron.tests.unit.extensions import test_quotasv2
from neutron.tests.unit import testlib_api
from testtools import matchers


class BaseTestCase(n_base.BaseTestCase):
    pass


class NeutronDbPluginV2TestCase(
    test_db_base_plugin_v2.NeutronDbPluginV2TestCase):

    def set_override(self, lbaas_provider):
        # override the default service provider
        self.service_providers = (
            mock.patch.object(st_db.ServiceTypeManager,
                              'get_service_providers').start())
        self.service_providers.return_value = (
            self._to_provider_dicts(lbaas_provider))

        # need to reload provider configuration
        st_db.ServiceTypeManager._instance = None

    def new_list_request(self, resource, fmt=None, params=None, id=None,
                         subresource=None):
        return self._req(
            'GET', resource, None, fmt, params=params, subresource=subresource,
            id=id
        )

    def new_show_request(self, resource, id, fmt=None,
                         subresource=None, sub_id=None, fields=None):
        if fields:
            params = "&".join(["fields=%s" % x for x in fields])
        else:
            params = None
        return self._req('GET', resource, None, fmt, id=id,
                         params=params, subresource=subresource, sub_id=sub_id)

    def new_update_request_lbaas(self, resource, data, id, fmt=None,
                           subresource=None, context=None, sub_id=None):
        return self._req(
            'PUT', resource, data, fmt, id=id, subresource=subresource,
            context=context, sub_id=sub_id
        )

    def _to_provider_dicts(self, lbaas_provider):
        provider_dicts = []
        for provider in lbaas_provider:
            bits = provider.split(':')
            p = {
                'service_type': bits[0],
                'name': bits[1],
                'driver': bits[2]
            }
            if len(bits) == 4:
                p['default'] = True
            provider_dicts.append(p)
        return provider_dicts

    def _test_list_with_sort(self, resource,
                             items, sorts, resources=None,
                             query_params='',
                             id=None,
                             subresource=None,
                             subresources=None):
        query_str = query_params
        for key, direction in sorts:
            query_str = query_str + "&sort_key=%s&sort_dir=%s" % (key,
                                                                  direction)
        if not resources:
            resources = '%ss' % resource
        if subresource and not subresources:
            subresources = '%ss' % subresource
        req = self.new_list_request(resources,
                                    params=query_str,
                                    id=id,
                                    subresource=subresources)
        api = self._api_for_resource(resources)
        res = self.deserialize(self.fmt, req.get_response(api))
        if subresource:
            resource = subresource
        if subresources:
            resources = subresources
        resource = resource.replace('-', '_')
        resources = resources.replace('-', '_')
        expected_res = [item[resource]['id'] for item in items]
        self.assertEqual(expected_res, [n['id'] for n in res[resources]])

    def _test_list_with_pagination(self, resource, items, sort,
                                   limit, expected_page_num,
                                   resources=None,
                                   query_params='',
                                   verify_key='id',
                                   id=None,
                                   subresource=None,
                                   subresources=None):
        if not resources:
            resources = '%ss' % resource
        if subresource and not subresources:
            subresources = '%ss' % subresource
        query_str = query_params + '&' if query_params else ''
        query_str = query_str + ("limit=%s&sort_key=%s&"
                                 "sort_dir=%s") % (limit, sort[0], sort[1])
        req = self.new_list_request(resources, params=query_str, id=id,
                                    subresource=subresources)
        items_res = []
        page_num = 0
        api = self._api_for_resource(resources)
        if subresource:
            resource = subresource
        if subresources:
            resources = subresources
        resource = resource.replace('-', '_')
        resources = resources.replace('-', '_')
        while req:
            page_num = page_num + 1
            res = self.deserialize(self.fmt, req.get_response(api))
            self.assertThat(len(res[resources]),
                            matchers.LessThan(limit + 1))
            items_res = items_res + res[resources]
            req = None
            if '%s_links' % resources in res:
                for link in res['%s_links' % resources]:
                    if link['rel'] == 'next':
                        content_type = 'application/%s' % self.fmt
                        req = testlib_api.create_request(link['href'],
                                                         '', content_type)
                        self.assertEqual(len(res[resources]),
                                         limit)
        self.assertEqual(expected_page_num, page_num)
        self.assertEqual([item[resource][verify_key] for item in items],
                         [n[verify_key] for n in items_res])

    def _test_list_with_pagination_reverse(self, resource, items, sort,
                                           limit, expected_page_num,
                                           resources=None,
                                           query_params='',
                                           id=None,
                                           subresource=None,
                                           subresources=None):
        if not resources:
            resources = '%ss' % resource
        if subresource and not subresources:
            subresources = '%ss' % subresource
        resource = resource.replace('-', '_')
        api = self._api_for_resource(resources)
        if subresource:
            marker = items[-1][subresource]['id']
        else:
            marker = items[-1][resource]['id']
        query_str = query_params + '&' if query_params else ''
        query_str = query_str + ("limit=%s&page_reverse=True&"
                                 "sort_key=%s&sort_dir=%s&"
                                 "marker=%s") % (limit, sort[0], sort[1],
                                                 marker)
        req = self.new_list_request(resources, params=query_str, id=id,
                                    subresource=subresources)
        if subresource:
            resource = subresource
        if subresources:
            resources = subresources
        item_res = [items[-1][resource]]
        page_num = 0
        resources = resources.replace('-', '_')
        while req:
            page_num = page_num + 1
            res = self.deserialize(self.fmt, req.get_response(api))
            self.assertThat(len(res[resources]),
                            matchers.LessThan(limit + 1))
            res[resources].reverse()
            item_res = item_res + res[resources]
            req = None
            if '%s_links' % resources in res:
                for link in res['%s_links' % resources]:
                    if link['rel'] == 'previous':
                        content_type = 'application/%s' % self.fmt
                        req = testlib_api.create_request(link['href'],
                                                         '', content_type)
                        self.assertEqual(len(res[resources]),
                                         limit)
        self.assertEqual(expected_page_num, page_num)
        expected_res = [item[resource]['id'] for item in items]
        expected_res.reverse()
        self.assertEqual(expected_res, [n['id'] for n in item_res])

    def _delete(self, collection, id,
                expected_code=webob.exc.HTTPNoContent.code,
                neutron_context=None, subresource=None, sub_id=None):
        req = self.new_delete_request(collection, id, subresource=subresource,
                                      sub_id=sub_id)
        if neutron_context:
            # create a specific auth context for this request
            req.environ['neutron.context'] = neutron_context
        res = req.get_response(self._api_for_resource(collection))
        self.assertEqual(res.status_int, expected_code)


class ExtensionTestCase(ext_base.ExtensionTestCase):
    pass


class QuotaExtensionTestCase(test_quotasv2.QuotaExtensionTestCase):

    def setUp(self):
        super(QuotaExtensionTestCase, self).setUp()
        resource_registry.register_resource_by_name('pool')
        resource_registry.register_resource_by_name('loadbalancer')
        resource_registry.register_resource_by_name('listener')
        resource_registry.register_resource_by_name('healthmonitor')
        resource_registry.register_resource_by_name('member')



# Copyright 2015, 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from six.moves.urllib import parse
from tempest.lib.common import rest_client


class PoolsClientJSON(rest_client.RestClient):
    """
    Test Pools API
    """

    def list_pools(self, params=None):
        """List all pools"""
        url = 'v2.0/lbaas/pools'
        if params:
            url = '{0}?{1}'.format(url, parse.urlencode(params))
        resp, body = self.get(url)
        body = jsonutils.loads(body)
        self.expected_success(200, resp.status)
        return rest_client.ResponseBodyList(resp, body['pools'])

    def get_pool(self, pool_id, params=None):
        """List details of a pool"""
        url = 'v2.0/lbaas/pools/{pool_id}'.format(pool_id=pool_id)
        if params:
            url = '{0}?{1}'.format(url, parse.urlencode(params))
        resp, body = self.get(url)
        body = jsonutils.loads(body)
        self.expected_success(200, resp.status)
        return rest_client.ResponseBody(resp, body['pool'])

    def create_pool(self, **kwargs):
        """Create a pool"""
        url = 'v2.0/lbaas/pools'
        post_body = jsonutils.dumps({'pool': kwargs})
        resp, body = self.post(url, post_body)
        body = jsonutils.loads(body)
        self.expected_success(201, resp.status)
        return rest_client.ResponseBody(resp, body['pool'])

    def update_pool(self, pool_id, **kwargs):
        """Update a pool"""
        url = 'v2.0/lbaas/pools/{pool_id}'.format(pool_id=pool_id)
        put_body = jsonutils.dumps({'pool': kwargs})
        resp, body = self.put(url, put_body)
        body = jsonutils.loads(body)
        self.expected_success(200, resp.status)
        return rest_client.ResponseBody(resp, body['pool'])

    def delete_pool(self, pool_id):
        """Delete Pool"""
        url = 'v2.0/lbaas/pools/{pool_id}'.format(pool_id=pool_id)
        resp, body = self.delete(url)
        self.expected_success(204, resp.status)
        return rest_client.ResponseBody(resp, body)



# Copyright (c) 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from neutron.db import agentschedulers_db
from neutron.db.models import agent as agents_db
from neutron_lib.db import model_base
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import joinedload

from neutron_lbaas.extensions import lbaas_agentschedulerv2
from neutron_lbaas.services.loadbalancer import constants as lb_const

LOG = logging.getLogger(__name__)


class LoadbalancerAgentBinding(model_base.BASEV2):
    """Represents binding between neutron loadbalancer and agents."""

    __tablename__ = "lbaas_loadbalanceragentbindings"

    loadbalancer_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("lbaas_loadbalancers.id", ondelete='CASCADE'),
        primary_key=True)
    agent = orm.relation(agents_db.Agent)
    agent_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("agents.id", ondelete='CASCADE'),
        nullable=False)


class LbaasAgentSchedulerDbMixin(agentschedulers_db.AgentSchedulerDbMixin,
                                 lbaas_agentschedulerv2
                                 .LbaasAgentSchedulerPluginBase):

    agent_notifiers = {}

    def get_agent_hosting_loadbalancer(self, context,
                                       loadbalancer_id, active=None):
        query = context.session.query(LoadbalancerAgentBinding)
        query = query.options(joinedload('agent'))
        binding = query.get(loadbalancer_id)

        if (binding and self.is_eligible_agent(
                active, binding.agent)):
            return {'agent': self._make_agent_dict(binding.agent)}

    def get_lbaas_agents(self, context, active=None, filters=None):
        query = context.session.query(agents_db.Agent)
        query = query.filter_by(agent_type=lb_const.AGENT_TYPE_LOADBALANCERV2)
        if active is not None:
            query = query.filter_by(admin_state_up=active)
        if filters:
            for key, value in filters.items():
                column = getattr(agents_db.Agent, key, None)
                if column:
                    query = query.filter(column.in_(value))

        return [agent
                for agent in query
                if self.is_eligible_agent(active, agent)]

    def list_loadbalancers_on_lbaas_agent(self, context, id):
        query = context.session.query(
            LoadbalancerAgentBinding.loadbalancer_id)
        query = query.filter_by(agent_id=id)
        loadbalancer_ids = [item[0] for item in query]
        if loadbalancer_ids:
            lbs = self.get_loadbalancers(context,
                                         filters={'id': loadbalancer_ids})
            return lbs
        return []

    def get_lbaas_agent_candidates(self, device_driver, active_agents):
        candidates = []
        for agent in active_agents:
            agent_conf = self.get_configuration_dict(agent)
            if device_driver in agent_conf['device_drivers']:
                candidates.append(agent)
        return candidates

    def get_down_loadbalancer_bindings(self, context, agent_dead_limit):
            cutoff = self.get_cutoff_time(agent_dead_limit)
            return (context.session.query(LoadbalancerAgentBinding).join(
                agents_db.Agent).filter(
                    agents_db.Agent.heartbeat_timestamp < cutoff,
                    agents_db.Agent.admin_state_up))

    def _unschedule_loadbalancer(self, context, loadbalancer_id, agent_id):
        with context.session.begin(subtransactions=True):
            query = context.session.query(LoadbalancerAgentBinding)
            query = query.filter(
                LoadbalancerAgentBinding.loadbalancer_id == loadbalancer_id,
                LoadbalancerAgentBinding.agent_id == agent_id)
            query.delete()


class ChanceScheduler(object):
    """Allocate a loadbalancer agent for a vip in a random way."""

    def schedule(self, plugin, context, loadbalancer, device_driver):
        """Schedule the load balancer to an active loadbalancer agent if there
        is no enabled agent hosting it.
        """
        with context.session.begin(subtransactions=True):
            lbaas_agent = plugin.db.get_agent_hosting_loadbalancer(
                context, loadbalancer.id)
            if lbaas_agent:
                LOG.debug('Load balancer %(loadbalancer_id)s '
                          'has already been hosted'
                          ' by lbaas agent %(agent_id)s',
                          {'loadbalancer_id': loadbalancer.id,
                           'agent_id': lbaas_agent['agent']['id']})
                return

            active_agents = plugin.db.get_lbaas_agents(context, active=True)
            if not active_agents:
                LOG.warning(
                    'No active lbaas agents for load balancer %s',
                    loadbalancer.id)
                return

            candidates = plugin.db.get_lbaas_agent_candidates(device_driver,
                                                              active_agents)
            if not candidates:
                LOG.warning('No lbaas agent supporting device driver %s',
                            device_driver)
                return

            chosen_agent = random.choice(candidates)
            binding = LoadbalancerAgentBinding()
            binding.agent = chosen_agent
            binding.loadbalancer_id = loadbalancer.id
            context.session.add(binding)
            LOG.debug(
                'Load balancer %(loadbalancer_id)s is scheduled '
                'to lbaas agent %(agent_id)s', {
                    'loadbalancer_id': loadbalancer.id,
                    'agent_id': chosen_agent['id']}
            )
            return chosen_agent



# Copyright 2014-2016 Rackspace US, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc

import six


@six.add_metaclass(abc.ABCMeta)
class BarbicanAuth(object):
    @abc.abstractmethod
    def get_barbican_client(self, project_id):
        """Creates a Barbican client object.

        :param project_id: Project ID that the request will be used for
        :return: a Barbican Client object
        :raises Exception: if the client cannot be created
        """



# Copyright 2015, Radware LTD. All rights reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import threading
import time

import netaddr
from neutron_lib import constants as n_constants
from neutron_lib import context
from neutron_lib.plugins import constants as pg_constants
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils
from six.moves import queue as Queue

import neutron_lbaas.common.cert_manager
from neutron_lbaas.drivers.radware import base_v2_driver
from neutron_lbaas.drivers.radware import exceptions as r_exc
from neutron_lbaas.drivers.radware import rest_client as rest

CERT_MANAGER_PLUGIN = neutron_lbaas.common.cert_manager.get_backend()
TEMPLATE_HEADER = {'Content-Type':
                   'application/vnd.com.radware.vdirect.'
                   'template-parameters+json'}
PROVISION_HEADER = {'Content-Type':
                    'application/vnd.com.radware.'
                    'vdirect.status+json'}
CREATE_SERVICE_HEADER = {'Content-Type':
                         'application/vnd.com.radware.'
                         'vdirect.adc-service-specification+json'}

PROPERTY_DEFAULTS = {'type': 'none',
                     'cookie_name': 'none',
                     'url_path': '/',
                     'http_method': 'GET',
                     'expected_codes': '200',
                     'subnet': '255.255.255.255',
                     'mask': '255.255.255.255',
                     'gw': '255.255.255.255',
                     }
LOADBALANCER_PROPERTIES = ['vip_address', 'admin_state_up']
LISTENER_PROPERTIES = ['id', 'protocol_port', 'protocol',
                       'connection_limit', 'admin_state_up']
DEFAULT_CERT_PROPERTIES = ['id', 'certificate', 'intermediates',
                           'private_key', 'passphrase']
SNI_CERT_PROPERTIES = DEFAULT_CERT_PROPERTIES + ['position']
L7_RULE_PROPERTIES = ['id', 'type', 'compare_type',
                      'key', 'value', 'admin_state_up']
L7_POLICY_PROPERTIES = ['id', 'action', 'redirect_pool_id',
                        'redirect_url', 'position', 'admin_state_up']
DEFAULT_POOL_PROPERTIES = ['id']
POOL_PROPERTIES = ['id', 'protocol', 'lb_algorithm', 'admin_state_up']
MEMBER_PROPERTIES = ['id', 'address', 'protocol_port', 'weight',
                     'admin_state_up', 'subnet', 'mask', 'gw']
SESSION_PERSISTENCY_PROPERTIES = ['type', 'cookie_name']
HEALTH_MONITOR_PROPERTIES = ['type', 'delay', 'timeout', 'max_retries',
                             'admin_state_up', 'url_path', 'http_method',
                             'expected_codes', 'id']

LOG = logging.getLogger(__name__)


class RadwareLBaaSV2Driver(base_v2_driver.RadwareLBaaSBaseV2Driver):
    #
    # Assumptions:
    # 1) We have only one worflow that takes care of l2-l4 and service creation
    # 2) The workflow template exists on the vDirect server
    # 3) The workflow expose one operation named 'update' (plus ctor and dtor)
    # 4) The 'update' operation gets the loadbalancer object graph as input
    # 5) The object graph is enehanced by our code before it is sent to the
    #    workflow
    # 6) Async operations are handled by a different thread
    #
    def __init__(self, plugin):
        base_v2_driver.RadwareLBaaSBaseV2Driver.__init__(self, plugin)
        rad = cfg.CONF.radwarev2
        rad_debug = cfg.CONF.radwarev2_debug
        self.plugin = plugin
        self.service = {
            "name": "_REPLACE_",
            "tenantId": "_REPLACE_",
            "haPair": rad.service_ha_pair,
            "sessionMirroringEnabled": rad.service_session_mirroring_enabled,
            "primary": {
                "capacity": {
                    "throughput": rad.service_throughput,
                    "sslThroughput": rad.service_ssl_throughput,
                    "compressionThroughput":
                    rad.service_compression_throughput,
                    "cache": rad.service_cache
                },
                "network": {
                    "type": "portgroup",
                    "portgroups": '_REPLACE_'
                },
                "adcType": rad.service_adc_type,
                "acceptableAdc": "Exact"
            }
        }
        if rad.service_resource_pool_ids:
            ids = rad.service_resource_pool_ids
            self.service['resourcePoolIds'] = [
                {'id': id} for id in ids
            ]
        else:
            self.service['resourcePoolIds'] = []

        if rad.service_isl_vlan:
            self.service['islVlan'] = rad.service_isl_vlan
        self.workflow_template_name = rad.workflow_template_name
        self.child_workflow_template_names = rad.child_workflow_template_names
        self.workflow_params = rad.workflow_params
        self.workflow_action_name = rad.workflow_action_name
        self.stats_action_name = rad.stats_action_name
        vdirect_address = rad.vdirect_address
        sec_server = rad.ha_secondary_address
        self.rest_client = rest.vDirectRESTClient(
            server=vdirect_address,
            secondary_server=sec_server,
            user=rad.vdirect_user,
            password=rad.vdirect_password,
            port=rad.port,
            ssl=rad.ssl,
            ssl_verify_context=rad.ssl_verify_context,
            timeout=rad.timeout,
            base_uri=rad.base_uri)
        self.workflow_params['provision_service'] = rad_debug.provision_service
        self.workflow_params['configure_l3'] = rad_debug.configure_l3
        self.workflow_params['configure_l4'] = rad_debug.configure_l4
        self.configure_allowed_address_pairs =\
            rad.configure_allowed_address_pairs

        self.queue = Queue.Queue()
        self.completion_handler = OperationCompletionHandler(self.queue,
                                                             self.rest_client,
                                                             plugin)
        self.workflow_templates_exists = False
        self.completion_handler.setDaemon(True)
        self.completion_handler_started = False

    def _start_completion_handling_thread(self):
        if not self.completion_handler_started:
            LOG.info('Starting operation completion handling thread')
            self.completion_handler.start()
            self.completion_handler_started = True

    @staticmethod
    def _get_wf_name(lb):
        return 'LB_' + lb.id

    @log_helpers.log_method_call
    def _verify_workflow_templates(self):
        """Verify the existence of workflows on vDirect server."""
        resource = '/api/workflowTemplate/'
        workflow_templates = {self.workflow_template_name: False}
        for child_wf_name in self.child_workflow_template_names:
            workflow_templates[child_wf_name] = False
        response = _rest_wrapper(self.rest_client.call('GET',
                                                       resource,
                                                       None,
                                                       None), [200])
        for workflow_template in workflow_templates.keys():
            for template in response:
                if workflow_template == template['name']:
                    workflow_templates[workflow_template] = True
                    break
        for template, found in workflow_templates.items():
            if not found:
                raise r_exc.WorkflowTemplateMissing(
                    workflow_template=template)

    @log_helpers.log_method_call
    def workflow_exists(self, lb):
        """Create workflow for loadbalancer instance"""
        wf_name = self._get_wf_name(lb)
        wf_resource = '/api/workflow/%s' % (wf_name)
        try:
            _rest_wrapper(self.rest_client.call(
                'GET', wf_resource, None, None),
                [200])
        except Exception:
            return False
        return True

    @log_helpers.log_method_call
    def _create_workflow(self, lb, lb_network_id, proxy_network_id):
        """Create workflow for loadbalancer instance"""

        self._verify_workflow_templates()

        wf_name = self._get_wf_name(lb)
        service = copy.deepcopy(self.service)
        service['tenantId'] = lb.tenant_id
        service['name'] = 'srv_' + lb_network_id

        if lb_network_id != proxy_network_id:
            self.workflow_params["twoleg_enabled"] = True
            service['primary']['network']['portgroups'] = [
                lb_network_id, proxy_network_id]
        else:
            self.workflow_params["twoleg_enabled"] = False
            service['primary']['network']['portgroups'] = [lb_network_id]

        tmpl_resource = '/api/workflowTemplate/%s?name=%s' % (
            self.workflow_template_name, wf_name)
        _rest_wrapper(self.rest_client.call(
            'POST', tmpl_resource,
            {'parameters': dict(self.workflow_params,
                                service_params=service),
            'tenants': [lb.tenant_id]},
            TEMPLATE_HEADER))

    @log_helpers.log_method_call
    def get_stats(self, ctx, lb):

        wf_name = self._get_wf_name(lb)
        resource = '/api/workflow/%s/action/%s' % (
            wf_name, self.stats_action_name)
        response = _rest_wrapper(self.rest_client.call('POST', resource,
                                 None, TEMPLATE_HEADER), success_codes=[202])
        LOG.debug('stats_action  response: %s ', response)

        resource = '/api/workflow/%s/parameters' % (wf_name)
        response = _rest_wrapper(self.rest_client.call('GET', resource,
                                 None, TEMPLATE_HEADER), success_codes=[200])
        LOG.debug('stats_values  response: %s ', response)
        return response['stats']

    @log_helpers.log_method_call
    def execute_workflow(self, ctx, manager, data_model,
                         old_data_model=None, delete=False):
        lb = data_model.root_loadbalancer

        # Get possible proxy subnet.
        # Proxy subnet equals to LB subnet if no proxy
        # is necessary.
        # Get subnet id of any member located on different than
        # loadbalancer's network. If returned subnet id is the subnet id
        # of loadbalancer - all members are accessible from loadbalancer's
        # network, meaning no second leg or static routes are required.
        # Otherwise, create proxy port on found member's subnet and get its
        # address as a proxy address for loadbalancer instance
        lb_subnet = self.plugin.db._core_plugin.get_subnet(
            ctx, lb.vip_subnet_id)
        proxy_subnet = lb_subnet
        proxy_port_address = lb.vip_address

        if not self.workflow_exists(lb):
            # Create proxy port if needed
            proxy_port_subnet_id = self._get_proxy_port_subnet_id(lb)
            if proxy_port_subnet_id != lb.vip_subnet_id:
                proxy_port = self._create_proxy_port(
                    ctx, lb, proxy_port_subnet_id)
                proxy_subnet = self.plugin.db._core_plugin.get_subnet(
                    ctx, proxy_port['subnet_id'])
                proxy_port_address = proxy_port['ip_address']

            self._create_workflow(lb,
                                  lb_subnet['network_id'],
                                  proxy_subnet['network_id'])
        else:
            # Check if proxy port exists
            proxy_port = self._get_proxy_port(ctx, lb)
            if proxy_port:
                proxy_subnet = self.plugin.db._core_plugin.get_subnet(
                    ctx, proxy_port['subnet_id'])
                proxy_port_address = proxy_port['ip_address']

        # Build objects graph
        objects_graph = self._build_objects_graph(ctx, lb, data_model,
                                                  proxy_port_address,
                                                  proxy_subnet,
                                                  delete)
        LOG.debug("Radware vDirect LB object graph is " + str(objects_graph))

        wf_name = self._get_wf_name(lb)
        resource = '/api/workflow/%s/action/%s' % (
            wf_name, self.workflow_action_name)
        response = _rest_wrapper(self.rest_client.call('POST', resource,
                                 {'parameters': objects_graph},
                                 TEMPLATE_HEADER), success_codes=[202])
        LOG.debug('_update_workflow response: %s ', response)

        oper = OperationAttributes(
            manager, response['uri'], lb,
            data_model, old_data_model,
            delete=delete)

        LOG.debug('Pushing operation %s to the queue', oper)
        self._start_completion_handling_thread()
        self.queue.put_nowait(oper)

    def remove_workflow(self, ctx, manager, lb):
        wf_name = self._get_wf_name(lb)
        LOG.debug('Remove the workflow %s', wf_name)
        resource = '/api/workflow/%s' % (wf_name)
        rest_return = self.rest_client.call('DELETE', resource, None, None)
        response = _rest_wrapper(rest_return, [204, 202, 404])
        if rest_return[rest.RESP_STATUS] in [404]:
            try:
                self._delete_proxy_port(ctx, lb)
                LOG.debug('Proxy port for LB %s was deleted', lb.id)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error('Proxy port deletion for LB %s failed', lb.id)
            manager.successful_completion(ctx, lb, delete=True)
        else:
            oper = OperationAttributes(
                manager, response['uri'], lb,
                lb, old_data_model=None,
                delete=True,
                post_operation_function=self._delete_proxy_port)

            self._start_completion_handling_thread()
            self.queue.put_nowait(oper)

    def _build_objects_graph(self, ctx, lb, data_model,
                             proxy_port_address, proxy_subnet,
                             deleted):
        """Iterate over the LB model starting from root lb entity
        and build its JSON representtaion for vDirect
        """
        deleted_ids = []
        if deleted:
            deleted_ids.append(data_model.id)

        graph = {}
        for prop in LOADBALANCER_PROPERTIES:
            graph[prop] = getattr(lb, prop, PROPERTY_DEFAULTS.get(prop))

        graph['pip_address'] = proxy_port_address
        graph['configure_allowed_address_pairs'] =\
            self.configure_allowed_address_pairs

        graph['listeners'] = []
        listeners = [
            listener for listener in lb.listeners
            if listener.provisioning_status != n_constants.PENDING_DELETE and
            (listener.default_pool and
             listener.default_pool.provisioning_status !=
             n_constants.PENDING_DELETE and
             listener.default_pool.id not in deleted_ids and
             listener.default_pool.members)]
        for listener in listeners:
            listener_dict = {}
            for prop in LISTENER_PROPERTIES:
                listener_dict[prop] = getattr(
                    listener, prop, PROPERTY_DEFAULTS.get(prop))

            cert_mgr = CERT_MANAGER_PLUGIN.CertManager()

            if listener.default_tls_container_id:
                default_cert = cert_mgr.get_cert(
                    project_id=listener.tenant_id,
                    cert_ref=listener.default_tls_container_id,
                    resource_ref=cert_mgr.get_service_url(
                        listener.loadbalancer_id),
                    service_name='Neutron LBaaS v2 Radware provider')
                def_cert_dict = {
                    'id': listener.default_tls_container_id,
                    'certificate': default_cert.get_certificate(),
                    'intermediates': default_cert.get_intermediates(),
                    'private_key': default_cert.get_private_key(),
                    'passphrase': default_cert.get_private_key_passphrase()}
                listener_dict['default_tls_certificate'] = def_cert_dict

            if listener.sni_containers:
                listener_dict['sni_tls_certificates'] = []
                for sni_container in listener.sni_containers:
                    sni_cert = cert_mgr.get_cert(
                        project_id=listener.tenant_id,
                        cert_ref=sni_container.tls_container_id,
                        resource_ref=cert_mgr.get_service_url(
                            listener.loadbalancer_id),
                        service_name='Neutron LBaaS v2 Radware provider')
                    listener_dict['sni_tls_certificates'].append(
                        {'id': sni_container.tls_container_id,
                         'position': sni_container.position,
                         'certificate': sni_cert.get_certificate(),
                         'intermediates': sni_cert.get_intermediates(),
                         'private_key': sni_cert.get_private_key(),
                         'passphrase': sni_cert.get_private_key_passphrase()})

            listener_dict['l7_policies'] = []
            policies = [
                policy for policy in listener.l7_policies
                if policy.provisioning_status != n_constants.PENDING_DELETE and
                policy.id not in deleted_ids]
            for policy in policies:
                policy_dict = {}
                for prop in L7_POLICY_PROPERTIES:
                    policy_dict[prop] = getattr(
                        policy, prop, PROPERTY_DEFAULTS.get(prop))
                policy_dict['rules'] = []
                rules = [
                    rule for rule in policy.rules
                    if rule.provisioning_status !=
                    n_constants.PENDING_DELETE and
                    rule.id not in deleted_ids]
                for rule in rules:
                    rule_dict = {}
                    for prop in L7_RULE_PROPERTIES:
                        rule_dict[prop] = getattr(
                            rule, prop, PROPERTY_DEFAULTS.get(prop))
                    policy_dict['rules'].append(rule_dict)
                if policy_dict['rules']:
                    listener_dict['l7_policies'].append(policy_dict)

            def_pool_dict = {'id': listener.default_pool.id}

            if listener.default_pool.session_persistence:
                sess_pers_dict = {}
                for prop in SESSION_PERSISTENCY_PROPERTIES:
                    sess_pers_dict[prop] = getattr(
                        listener.default_pool.session_persistence, prop,
                        PROPERTY_DEFAULTS.get(prop))
                def_pool_dict['sessionpersistence'] = sess_pers_dict
            listener_dict['default_pool'] = def_pool_dict

            graph['listeners'].append(listener_dict)

        graph['pools'] = []
        pools = [
            pool for pool in lb.pools
            if pool.provisioning_status != n_constants.PENDING_DELETE and
            pool.id not in deleted_ids]
        for pool in pools:
            pool_dict = {}
            for prop in POOL_PROPERTIES:
                pool_dict[prop] = getattr(
                    pool, prop,
                    PROPERTY_DEFAULTS.get(prop))

            if (pool.healthmonitor and
                pool.healthmonitor.provisioning_status !=
                n_constants.PENDING_DELETE and
                pool.healthmonitor.id not in deleted_ids):
                hm_dict = {}
                for prop in HEALTH_MONITOR_PROPERTIES:
                    hm_dict[prop] = getattr(
                        pool.healthmonitor, prop,
                        PROPERTY_DEFAULTS.get(prop))
                pool_dict['healthmonitor'] = hm_dict

            pool_dict['members'] = []
            members = [
                member for member in pool.members
                if member.provisioning_status != n_constants.PENDING_DELETE and
                member.id not in deleted_ids]
            for member in members:
                member_dict = {}
                for prop in MEMBER_PROPERTIES:
                    member_dict[prop] = getattr(
                        member, prop,
                        PROPERTY_DEFAULTS.get(prop))
                if (proxy_port_address != lb.vip_address and
                    netaddr.IPAddress(member.address)
                    not in netaddr.IPNetwork(proxy_subnet['cidr'])):
                    self._accomplish_member_static_route_data(
                        ctx, member, member_dict,
                        proxy_subnet['gateway_ip'])
                pool_dict['members'].append(member_dict)
            graph['pools'].append(pool_dict)

        return graph

    def _get_proxy_port_subnet_id(self, lb):
        """Look for at least one member of any listener's pool
        that is located on subnet different than loabalancer's subnet.
        If such member found, return its subnet id.
        Otherwise, return loadbalancer's subnet id
        """
        for listener in lb.listeners:
            if listener.default_pool:
                for member in listener.default_pool.members:
                    if lb.vip_subnet_id != member.subnet_id:
                        return member.subnet_id
        return lb.vip_subnet_id

    def _create_proxy_port(self,
        ctx, lb, proxy_port_subnet_id):
        """Check if proxy port was created earlier.
        If not, create a new port on proxy subnet and return its ip address.
        Returns port IP address
        """
        proxy_port = self._get_proxy_port(ctx, lb)
        if proxy_port:
            LOG.info('LB %(lb_id)s proxy port exists on subnet '
                     '%(subnet_id)s with ip address %(ip_address)s' %
                     {'lb_id': lb.id, 'subnet_id': proxy_port['subnet_id'],
                      'ip_address': proxy_port['ip_address']})
            return proxy_port

        proxy_port_name = 'proxy_' + lb.id
        proxy_port_subnet = self.plugin.db._core_plugin.get_subnet(
            ctx, proxy_port_subnet_id)
        proxy_port_data = {
            'tenant_id': lb.tenant_id,
            'name': proxy_port_name,
            'network_id': proxy_port_subnet['network_id'],
            'mac_address': n_constants.ATTR_NOT_SPECIFIED,
            'admin_state_up': False,
            'device_id': '',
            'device_owner': 'neutron:' + pg_constants.LOADBALANCERV2,
            'fixed_ips': [{'subnet_id': proxy_port_subnet_id}]
        }
        proxy_port = self.plugin.db._core_plugin.create_port(
            ctx, {'port': proxy_port_data})
        proxy_port_ip_data = proxy_port['fixed_ips'][0]

        LOG.info('LB %(lb_id)s proxy port created on subnet %(subnet_id)s '
                 'with ip address %(ip_address)s' %
                 {'lb_id': lb.id, 'subnet_id': proxy_port_ip_data['subnet_id'],
                  'ip_address': proxy_port_ip_data['ip_address']})

        return proxy_port_ip_data

    def _get_proxy_port(self, ctx, lb):
        ports = self.plugin.db._core_plugin.get_ports(
            ctx, filters={'name': ['proxy_' + lb.id], })
        if not ports:
            return None

        proxy_port = ports[0]
        return proxy_port['fixed_ips'][0]

    def _delete_proxy_port(self, ctx, lb):
        port_filter = {
            'name': ['proxy_' + lb.id],
        }
        ports = self.plugin.db._core_plugin.get_ports(
            ctx, filters=port_filter)
        if ports:
            proxy_port = ports[0]
            proxy_port_ip_data = proxy_port['fixed_ips'][0]
            try:
                LOG.info('Deleting LB %(lb_id)s proxy port on subnet  '
                         '%(subnet_id)s with ip address %(ip_address)s' %
                         {'lb_id': lb.id,
                          'subnet_id': proxy_port_ip_data['subnet_id'],
                          'ip_address': proxy_port_ip_data['ip_address']})
                self.plugin.db._core_plugin.delete_port(
                    ctx, proxy_port['id'])

            except Exception as exception:
                # stop exception propagation, nport may have
                # been deleted by other means
                LOG.warning('Proxy port deletion failed: %r',
                            exception)

    def _accomplish_member_static_route_data(self,
        ctx, member, member_data, proxy_gateway_ip):
        member_ports = self.plugin.db._core_plugin.get_ports(
            ctx,
            filters={'fixed_ips': {'ip_address': [member.address]},
                     'tenant_id': [member.tenant_id]})
        if len(member_ports) == 1:
            member_port = member_ports[0]
            member_port_ip_data = member_port['fixed_ips'][0]
            LOG.debug('member_port_ip_data:' + repr(member_port_ip_data))
            member_subnet = self.plugin.db._core_plugin.get_subnet(
                ctx,
                member_port_ip_data['subnet_id'])
            LOG.debug('member_subnet:' + repr(member_subnet))
            member_network = netaddr.IPNetwork(member_subnet['cidr'])
            member_data['subnet'] = str(member_network.network)
            member_data['mask'] = str(member_network.netmask)
        else:
            member_data['subnet'] = member_data['address']
        member_data['gw'] = proxy_gateway_ip


class OperationCompletionHandler(threading.Thread):

    """Update DB with operation status or delete the entity from DB."""

    def __init__(self, queue, rest_client, plugin):
        threading.Thread.__init__(self)
        self.queue = queue
        self.rest_client = rest_client
        self.plugin = plugin
        self.stoprequest = threading.Event()
        self.opers_to_handle_before_rest = 0

    def join(self, timeout=None):
        self.stoprequest.set()
        super(OperationCompletionHandler, self).join(timeout)

    def handle_operation_completion(self, oper):
        result = self.rest_client.call('GET',
                                       oper.operation_url,
                                       None,
                                       None)
        LOG.debug('Operation completion requested %(uri)s and got: %(result)s',
                  {'uri': oper.operation_url, 'result': result})
        completed = result[rest.RESP_DATA]['complete']
        reason = result[rest.RESP_REASON],
        description = result[rest.RESP_STR]
        if completed:
            # operation is done - update the DB with the status
            # or delete the entire graph from DB
            success = result[rest.RESP_DATA]['success']
            sec_to_completion = time.time() - oper.creation_time
            debug_data = {'oper': oper,
                          'sec_to_completion': sec_to_completion,
                          'success': success}
            LOG.debug('Operation %(oper)s is completed after '
                      '%(sec_to_completion)d sec '
                      'with success status: %(success)s :',
                      debug_data)
            if not success:
                # failure - log it and set the return ERROR as DB state
                if reason or description:
                    msg = 'Reason:%s. Description:%s' % (reason, description)
                else:
                    msg = "unknown"
                error_params = {"operation": oper, "msg": msg}
                LOG.error(
                    'Operation %(operation)s failed. Reason: %(msg)s',
                    error_params)
                oper.status = n_constants.ERROR
                OperationCompletionHandler._run_post_failure_function(oper)
            else:
                oper.status = n_constants.ACTIVE
                OperationCompletionHandler._run_post_success_function(oper)

        return completed

    def run(self):
        while not self.stoprequest.isSet():
            try:
                oper = self.queue.get(timeout=1)

                # Get the current queue size (N) and set the counter with it.
                # Handle N operations with no intermission.
                # Once N operations handles, get the size again and repeat.
                if self.opers_to_handle_before_rest <= 0:
                    self.opers_to_handle_before_rest = self.queue.qsize() + 1

                LOG.debug('Operation consumed from the queue: ' +
                          str(oper))
                # check the status - if oper is done: update the db ,
                # else push the oper again to the queue
                if not self.handle_operation_completion(oper):
                    LOG.debug('Operation %s is not completed yet..' % oper)
                    # Not completed - push to the queue again
                    self.queue.put_nowait(oper)

                self.queue.task_done()
                self.opers_to_handle_before_rest -= 1

                # Take one second rest before start handling
                # new operations or operations handled before
                if self.opers_to_handle_before_rest <= 0:
                    time.sleep(1)

            except Queue.Empty:
                continue
            except Exception:
                LOG.error(
                    "Exception was thrown inside OperationCompletionHandler")

    @staticmethod
    def _run_post_success_function(oper):
        try:
            ctx = context.get_admin_context()
            if oper.post_operation_function:
                oper.post_operation_function(ctx, oper.data_model)
            oper.manager.successful_completion(ctx, oper.data_model,
                                               delete=oper.delete)
            LOG.debug('Post-operation success function completed '
                      'for operation %s',
                      repr(oper))
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error('Post-operation success function failed '
                          'for operation %s',
                          repr(oper))

    @staticmethod
    def _run_post_failure_function(oper):
        try:
            ctx = context.get_admin_context()
            oper.manager.failed_completion(ctx, oper.data_model)
            LOG.debug('Post-operation failure function completed '
                      'for operation %s',
                      repr(oper))
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error('Post-operation failure function failed '
                          'for operation %s',
                          repr(oper))


class OperationAttributes(object):

    """Holds operation attributes"""

    def __init__(self,
                 manager,
                 operation_url,
                 lb,
                 data_model=None,
                 old_data_model=None,
                 delete=False,
                 post_operation_function=None):
        self.manager = manager
        self.operation_url = operation_url
        self.lb = lb
        self.data_model = data_model
        self.old_data_model = old_data_model
        self.delete = delete
        self.post_operation_function = post_operation_function
        self.creation_time = time.time()

    def __repr__(self):
        attrs = self.__dict__
        items = ("%s = %r" % (k, v) for k, v in attrs.items())
        return "<%s: {%s}>" % (self.__class__.__name__, ', '.join(items))


def _rest_wrapper(response, success_codes=None):
    """Wrap a REST call and make sure a valido status is returned."""
    success_codes = success_codes or [202]
    if not response:
        raise r_exc.RESTRequestFailure(
            status=-1,
            reason="Unknown",
            description="Unknown",
            success_codes=success_codes
        )
    elif response[rest.RESP_STATUS] not in success_codes:
        raise r_exc.RESTRequestFailure(
            status=response[rest.RESP_STATUS],
            reason=response[rest.RESP_REASON],
            description=response[rest.RESP_STR],
            success_codes=success_codes
        )
    else:
        LOG.debug("this is a respone: %s" % (response,))
        return response[rest.RESP_DATA]



#  Copyright 2015, Shane McGough, KEMPtechnologies
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_config import cfg

from neutron_lbaas._i18n import _


KEMP_OPTS = [
    cfg.StrOpt('lm_address', default='192.168.0.1',
               help=_('Management address of the LoadMaster appliance.')),
    cfg.StrOpt('lm_username', default='bal',
               help=_('The management user. Default is bal.')),
    cfg.StrOpt('lm_password', default='1fourall', secret=True,
               help=_('Password for management user. Default is 1fourall.')),
    cfg.IntOpt('check_interval', default=9,
               help=_('The interval between real server health checks.')),
    cfg.IntOpt('connect_timeout', default=4,
               help=_('The time to wait for a real server to respond to a '
                      'health check request.')),
    cfg.IntOpt('retry_count', default=2,
               help=_('If a real server fails to respond to a health check '
                      'request. The LoadMaster will retry the specified '
                      'number of times.')),
]



# Copyright 2016 A10 Networks
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import extensions
from neutron_lib import constants as n_constants
from neutron_lib import exceptions as nexception

from neutron_lbaas._i18n import _

EXTENDED_ATTRIBUTES_2_0 = {
    'loadbalancers': {
        'vip_subnet_id': {'allow_post': True, 'allow_put': False,
                          'validate': {'type:uuid': None},
                          'is_visible': True,
                          'default': n_constants.ATTR_NOT_SPECIFIED},
        'vip_network_id': {'allow_post': True, 'allow_put': False,
                           'validate': {'type:uuid': None},
                           'is_visible': False,
                           'default': n_constants.ATTR_NOT_SPECIFIED}
    }
}


class VipNetworkInvalid(nexception.BadRequest):
    message = _("VIP network %(network)s is invalid. "
                "There is no subnet in VIP network specified.")


class Lb_network_vip(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Create loadbalancer with network_id"

    @classmethod
    def get_alias(cls):
        return "lb_network_vip"

    @classmethod
    def get_description(cls):
        return "Create loadbalancer with network_id"

    @classmethod
    def get_updated(cls):
        return "2016-09-09T22:00:00-00:00"

    def get_required_extensions(self):
        return ["lbaasv2"]

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base


PROTOCOL_PORT = 80


class TestPools(base.BaseAdminTestCase):

    """
    Tests the following operations in the Neutron-LBaaS API using the
    REST client for Pools:

        list pools
        create pool
        get pool
        update pool
        delete pool
    """

    @classmethod
    def resource_setup(cls):
        super(TestPools, cls).resource_setup()
        if not utils.is_extension_enabled('lbaasv2', 'network'):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.load_balancer = cls._create_load_balancer(
            tenant_id=cls.subnet.get('tenant_id'),
            vip_subnet_id=cls.subnet.get('id'))

    def increment_protocol_port(self):
        global PROTOCOL_PORT
        PROTOCOL_PORT += 1

    def _prepare_and_create_pool(self, protocol=None, lb_algorithm=None,
                                 listener_id=None, cleanup=True, **kwargs):
        self.increment_protocol_port()
        if not protocol:
            protocol = self.pool_protocol
        if not lb_algorithm:
            lb_algorithm = 'ROUND_ROBIN'
        if not listener_id:
            listener = self._create_listener(
                loadbalancer_id=self.load_balancer.get('id'),
                protocol=self.listener_protocol,
                protocol_port=PROTOCOL_PORT, **kwargs)
            listener_id = listener.get('id')
        response = self._create_pool(protocol=protocol,
                                     lb_algorithm=lb_algorithm,
                                     listener_id=listener_id,
                                     **kwargs)
        if cleanup:
            self.addCleanup(self._delete_pool, response['id'])
        return response

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_pool_using_empty_tenant_field(self):
        """Test create pool with empty tenant field should fail"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          tenant_id="",
                          lb_algorithm='ROUND_ROBIN')

    @decorators.skip_because(bug="1468457")
    def test_create_pool_missing_tenant_id_for_other_tenant(self):
        """
        Test create pool with a missing tenant id field. Verify
        tenant_id does not match when creating pool vs.
        pool (admin client)
        """
        new_pool = self._prepare_and_create_pool()
        pool = self.pools_client.get_pool(new_pool.get('id'))
        pool_tenant = pool['tenant_id']
        self.assertNotEqual(pool_tenant, self.subnet['tenant_id'])

    @decorators.skip_because(bug="1468457")
    def test_create_pool_missing_tenant_id_for_admin(self):
        """
        Test create pool with a missing tenant id field. Verify
        tenant_id matches when creating pool vs. pool (admin client)
        """
        new_pool = self._prepare_and_create_pool()
        pool = self.pools_client.get_pool(new_pool.get('id'))
        pool_tenant = pool['tenant_id']
        self.assertEqual(pool_tenant, pool.get('tenant_id'))

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='smoke')
    def test_create_pool_for_another_tenant(self):
        """Test create pool for other tenant field"""
        tenant = 'deffb4d7c0584e89a8ec99551565713c'
        new_pool = self._prepare_and_create_pool(
            tenant_id=tenant)
        pool = self.pools_client.get_pool(new_pool.get('id'))
        pool_tenant = pool.get('tenant_id')
        self.assertEqual(pool_tenant, tenant)

    @decorators.attr(type='smoke')
    def test_update_pool_sesssion_persistence_app_cookie(self):
        """Test update admin pool's session persistence"""
        new_pool = self._prepare_and_create_pool()
        session_persistence = {"type": "APP_COOKIE",
                               "cookie_name": "my_cookie"}
        self._update_pool(new_pool.get('id'),
                          session_persistence=session_persistence)
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self.assertEqual(session_persistence, pool.get('session_persistence'))

    def test_update_pool_sesssion_persistence_app_to_http(self):
        """
        Test update admin pool's session persistence type from
        app cookie to http cookie
        """
        new_pool = self._prepare_and_create_pool()
        session_persistence = {"type": "APP_COOKIE",
                               "cookie_name": "my_cookie"}
        self._update_pool(new_pool.get('id'),
                          session_persistence=session_persistence)
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self.assertEqual(session_persistence, pool.get('session_persistence'))
        self._update_pool(new_pool.get('id'),
                          session_persistence={"type": "HTTP_COOKIE"})
        pool = self.pools_client.get_pool(new_pool.get('id'))
        session_persistence = {"type": "HTTP_COOKIE",
                               "cookie_name": None}
        self.assertEqual(session_persistence, pool.get('session_persistence'))

    @decorators.attr(type='smoke')
    def test_delete_pool(self):
        """Test delete admin pool"""
        new_pool = self._prepare_and_create_pool(cleanup=False)
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self._test_provisioning_status_if_exists(new_pool, pool)
        self.assertEqual(new_pool, pool)
        self._delete_pool(new_pool.get('id'))
        self.assertRaises(ex.NotFound, self.pools_client.get_pool,
                          new_pool.get('id'))



# Copyright 2015 Radware LTD. All rights reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import copy
import re

import mock
from neutron_lib import context
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_serialization import jsonutils
from six.moves import queue as Queue

from neutron_lbaas.common.cert_manager import cert_manager
from neutron_lbaas.drivers.radware import exceptions as r_exc
from neutron_lbaas.drivers.radware import v2_driver
from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.services.loadbalancer import constants as lb_con
from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2

GET_200 = ('/api/workflow/', '/api/workflowTemplate')
SERVER_DOWN_CODES = (-1, 301, 307)


class QueueMock(Queue.Queue):
    def __init__(self, completion_handler):
        self.completion_handler = completion_handler
        super(QueueMock, self).__init__()

    def put_nowait(self, oper):
        self.completion_handler(oper)


def _recover_function_mock(action, resource, data, headers, binary=False):
    pass


def rest_call_function_mock(action, resource, data, headers, binary=False):
    if rest_call_function_mock.RESPOND_WITH_ERROR:
        return 400, 'error_status', 'error_description', None
    if rest_call_function_mock.RESPOND_WITH_SERVER_DOWN in SERVER_DOWN_CODES:
        val = rest_call_function_mock.RESPOND_WITH_SERVER_DOWN
        return val, 'error_status', 'error_description', None
    if action == 'GET':
        return _get_handler(resource)
    elif action == 'DELETE':
        return _delete_handler(resource)
    elif action == 'POST':
        return _post_handler(resource, binary)
    else:
        return 0, None, None, None


def _get_handler(resource):
    if resource.startswith(GET_200[1]):
        return 200, '', '', rest_call_function_mock.WF_TEMPLATES_TO_RETURN

    if resource.startswith(GET_200[0]):
        if rest_call_function_mock.WORKFLOW_MISSING:
            data = jsonutils.loads('{"complete":"True", "success": "True"}')
            return 404, '', '', data
        elif resource.endswith('parameters'):
            return 200, '', '', {'stats': {'bytes_in': 100,
                'total_connections': 2, 'active_connections': 1,
                'bytes_out': 200}}
        else:
            return 200, '', '', ''

    if resource.startswith(GET_200):
        return 200, '', '', ''
    else:
        data = jsonutils.loads('{"complete":"True", "success": "True"}')
        return 202, '', '', data


def _delete_handler(resource):
    return 404, '', '', {'message': 'Not Found'}


def _post_handler(resource, binary):
    if re.search(r'/api/workflow/.+/action/.+', resource):
        data = jsonutils.loads('{"uri":"some_uri"}')
        return 202, '', '', data
    elif re.search(r'/api/service\?name=.+', resource):
        data = jsonutils.loads('{"links":{"actions":{"provision":"someuri"}}}')
        return 201, '', '', data
    elif binary:
        return 201, '', '', ''
    else:
        return 202, '', '', ''

RADWARE_PROVIDER = ('LOADBALANCERV2:radwarev2:neutron_lbaas.'
                    'drivers.radware.v2_driver.'
                    'RadwareLBaaSV2Driver:default')

WF_SRV_PARAMS = {
    "name": "_REPLACE_", "tenantId": "_REPLACE_", "haPair": False,
    "sessionMirroringEnabled": False, "islVlan": -1,
    "primary": {
        "capacity": {
            "throughput": 1000, "sslThroughput": 100,
            "compressionThroughput": 100, "cache": 20},
        "network": {
            "type": "portgroup", "portgroups": "_REPLACE_"},
        "adcType": "VA", "acceptableAdc": "Exact"},
    "resourcePoolIds": []}

WF_CREATE_PARAMS = {'parameters':
    {"provision_service": True, "configure_l3": True, "configure_l4": True,
     "twoleg_enabled": False, "ha_network_name": "HA-Network",
     "ha_ip_pool_name": "default", "allocate_ha_vrrp": True,
     "allocate_ha_ips": True, "data_port": 1,
     "data_ip_address": "192.168.200.99", "data_ip_mask": "255.255.255.0",
     "gateway": "192.168.200.1", "ha_port": 2}, 'tenants': "_REPLACE_"}

WF_APPLY_PARAMS = {
    'parameters': {'listeners': [], 'pools': [], 'admin_state_up': True,
    'configure_allowed_address_pairs': False,
    'pip_address': u'10.0.0.2', 'vip_address': u'10.0.0.2'}}

LISTENER = {
    'id': None,
    'admin_state_up': True,
    'protocol_port': 80,
    'protocol': lb_con.PROTOCOL_HTTP,
    'default_pool': None,
    'connection_limit': -1,
    'l7_policies': []}

L7_POLICY = {
    'id': None,
    'rules': [],
    'redirect_pool_id': None,
    'redirect_url': None,
    'action': lb_con.L7_POLICY_ACTION_REJECT,
    'position': 1,
    'admin_state_up': True}

L7_RULE = {
    'id': None,
    'type': lb_con.L7_RULE_TYPE_HOST_NAME,
    'compare_type': lb_con.L7_RULE_COMPARE_TYPE_EQUAL_TO,
    'admin_state_up': True,
    'key': None,
    'value': u'val1'}

DEFAULT_POOL = {'id': None}

SESSION_PERSISTENCE = {
    'type': 'APP_COOKIE',
    'cookie_name': 'sessionId'}

CERTIFICATE = {
    'id': None,
    'certificate': 'certificate',
    'intermediates': 'intermediates',
    'private_key': 'private_key',
    'passphrase': 'private_key_passphrase'}

SNI_CERTIFICATE = {
    'id': None,
    'position': 0,
    'certificate': 'certificate',
    'intermediates': 'intermediates',
    'private_key': 'private_key',
    'passphrase': 'private_key_passphrase'}

POOL = {
    'id': None,
    'protocol': lb_con.PROTOCOL_HTTP,
    'lb_algorithm': 'ROUND_ROBIN',
    'admin_state_up': True,
    'members': []}

MEMBER = {
    'id': None,
    'address': '10.0.1.10',
    'protocol_port': 80,
    'weight': 1, 'admin_state_up': True,
    'subnet': '255.255.255.255',
    'mask': '255.255.255.255',
    'gw': '255.255.255.255',
    'admin_state_up': True}

HM = {
    'id': None,
    'expected_codes': '200',
    'type': 'HTTP',
    'delay': 1,
    'timeout': 1,
    'max_retries': 1,
    'admin_state_up': True,
    'url_path': '/',
    'http_method': 'GET'}


class TestLBaaSDriverBase(
    test_db_loadbalancerv2.LbaasPluginDbTestCase):

    def setUp(self):
        super(TestLBaaSDriverBase, self).setUp(
            lbaas_provider=RADWARE_PROVIDER)

        self.plugin_instance = directory.get_plugin(constants.LOADBALANCERV2)
        self.driver = self.plugin_instance.drivers['radwarev2']


class TestLBaaSDriverRestClient(TestLBaaSDriverBase):
    def setUp(self):

        cfg.CONF.set_override('vdirect_address', '1.1.1.1',
                              group='radwarev2')
        cfg.CONF.set_override('ha_secondary_address', '1.1.1.2',
                              group='radwarev2')
        super(TestLBaaSDriverRestClient, self).setUp()

        self.flip_servers_mock = mock.Mock(
            return_value=None)
        self.recover_mock = mock.Mock(
            side_effect=_recover_function_mock)

        self.orig_recover = self.driver.rest_client._recover
        self.orig_flip_servers = self.driver.rest_client._flip_servers
        self.driver.rest_client._flip_servers = self.flip_servers_mock
        self.driver.rest_client._recover = self.recover_mock

    def test_recover_was_called(self):
        """Call REST client which fails and verify _recover is called."""
        self.driver.rest_client.call('GET', '/api/workflowTemplate',
                                     None, None)
        self.recover_mock.assert_called_once_with('GET',
                                                  '/api/workflowTemplate',
                                                  None, None, False)

    def test_flip_servers(self):
        server = self.driver.rest_client.server
        sec_server = self.driver.rest_client.secondary_server

        self.driver.rest_client._recover = self.orig_recover
        self.driver.rest_client._flip_servers = self.orig_flip_servers
        self.driver.rest_client.call('GET', '/api/workflowTemplate',
                                     None, None)
        self.assertEqual(server, self.driver.rest_client.secondary_server)
        self.assertEqual(sec_server, self.driver.rest_client.server)


class CertMock(cert_manager.Cert):
    def __init__(self, cert_container):
        pass

    def get_certificate(self):
        return "certificate"

    def get_intermediates(self):
        return "intermediates"

    def get_private_key(self):
        return "private_key"

    def get_private_key_passphrase(self):
        return "private_key_passphrase"


class TestLBaaSDriver(TestLBaaSDriverBase):

    @contextlib.contextmanager
    def loadbalancer(self, fmt=None, subnet=None, no_delete=False, **kwargs):
        with super(TestLBaaSDriver, self).loadbalancer(
            fmt, subnet, no_delete,
            vip_address=WF_APPLY_PARAMS['parameters']['vip_address'],
            **kwargs) as lb:
            self.wf_srv_params['name'] = 'srv_' + (
                subnet['subnet']['network_id'])
            self.wf_srv_params['tenantId'] = self._tenant_id
            self.wf_srv_params['primary']['network']['portgroups'] =\
                [subnet['subnet']['network_id']]
            self.wf_create_params['parameters']['service_params'] =\
                self.wf_srv_params
            yield lb

    @contextlib.contextmanager
    def listener(self, fmt=None, protocol='HTTP', loadbalancer_id=None,
                 protocol_port=80, default_pool_id=None, no_delete=False,
                 **kwargs):
        with super(TestLBaaSDriver, self).listener(
            fmt, protocol, loadbalancer_id, protocol_port, default_pool_id,
            no_delete, **kwargs) as listener:

            l = copy.deepcopy(LISTENER)
            self._update_dict(l, listener['listener'])
            if 'default_tls_container_ref' in kwargs:
                c = copy.deepcopy(CERTIFICATE)
                self._update_dict(c,
                                  {'id': kwargs['default_tls_container_ref']})
                l['default_tls_certificate'] = c
            if 'sni_container_refs' in kwargs:
                l['sni_tls_certificates'] = []
                pos = 0
                for ref in kwargs['sni_container_refs']:
                    s = copy.deepcopy(SNI_CERTIFICATE)
                    self._update_dict(s, {'id': ref, 'position': pos})
                    l['sni_tls_certificates'].append(s)
                    pos = pos + 1
            self.wf_apply_params['parameters']['listeners'].append(l)
            yield listener

    @contextlib.contextmanager
    def l7policy(self, listener_id, fmt=None,
                 action=lb_con.L7_POLICY_ACTION_REJECT,
                 no_delete=False, **kwargs):
        with super(TestLBaaSDriver, self).l7policy(
            listener_id, fmt, action, no_delete, **kwargs) as policy:

            p = copy.deepcopy(L7_POLICY)
            self._update_dict(p, policy['l7policy'])
            for l in self.wf_apply_params['parameters']['listeners']:
                if l['id'] == listener_id:
                    l['l7_policies'].append(p)
                    break
            yield policy

    @contextlib.contextmanager
    def l7policy_rule(self, l7policy_id, fmt=None, value='value1',
                      type=lb_con.L7_RULE_TYPE_HOST_NAME,
                      compare_type=lb_con.L7_RULE_COMPARE_TYPE_EQUAL_TO,
                      no_delete=False, **kwargs):
        with super(TestLBaaSDriver, self).l7policy_rule(
            l7policy_id, fmt, value, type, compare_type,
            no_delete, **kwargs) as rule:

            r = copy.deepcopy(L7_RULE)
            self._update_dict(r, rule['rule'])
            for l in self.wf_apply_params['parameters']['listeners']:
                #if l['id'] == listener_id:
                for p in l['l7_policies']:
                    if p['id'] == l7policy_id:
                        p['rules'].append(r)
                        break
            yield rule

    @contextlib.contextmanager
    def pool(self, fmt=None, protocol='HTTP', lb_algorithm='ROUND_ROBIN',
             no_delete=False, listener_id=None,
             loadbalancer_id=None, **kwargs):
        with super(TestLBaaSDriver, self).pool(
            fmt, protocol, lb_algorithm, no_delete, listener_id,
            loadbalancer_id, **kwargs) as pool:

            p = copy.deepcopy(POOL)
            self._update_dict(p, pool['pool'])
            self.wf_apply_params['parameters']['pools'].append(p)
            if listener_id:
                p = copy.deepcopy(DEFAULT_POOL)
                self._update_dict(p, pool['pool'])
                for l in self.wf_apply_params['parameters']['listeners']:
                    if l['id'] == listener_id:
                        l['default_pool'] = p
                        break
            if 'session_persistence' in kwargs:
                s = copy.deepcopy(SESSION_PERSISTENCE)
                self._update_dict(s, kwargs['session_persistence'])
                l['default_pool']['sessionpersistence'] = s
            yield pool

    @contextlib.contextmanager
    def member(self, fmt=None, pool_id='pool1id', address='127.0.0.1',
               protocol_port=80, subnet=None, no_delete=False,
               **kwargs):
        with super(TestLBaaSDriver, self).member(
            fmt, pool_id, address, protocol_port, subnet,
            no_delete, **kwargs) as member:

            m = copy.deepcopy(MEMBER)
            self._update_dict(m, member['member'])
            for p in self.wf_apply_params['parameters']['pools']:
                if p['id'] == pool_id:
                    p['members'].append(m)
                    break
            yield member

    @contextlib.contextmanager
    def healthmonitor(self, fmt=None, pool_id='pool1id', type='TCP', delay=1,
                      timeout=1, max_retries=1, no_delete=False, **kwargs):
        with super(TestLBaaSDriver, self).healthmonitor(
            fmt, pool_id, type, delay, timeout, max_retries,
            no_delete, **kwargs) as hm:

            m = copy.deepcopy(HM)
            self._update_dict(m, hm['healthmonitor'])
            for p in self.wf_apply_params['parameters']['pools']:
                if p['id'] == pool_id:
                    p['healthmonitor'] = m
                    break
            yield hm

    def _update_dict(self, o, d):
        for p in list(o.keys()):
            if p in d:
                o[p] = d[p]

    def update_member(self, pool_id, **kwargs):
        m = copy.deepcopy(MEMBER)
        self._update_dict(m, kwargs)
        for p in self.wf_apply_params['parameters']['pools']:
            if p['id'] == pool_id:
                for mem in p['members']:
                    if mem['id'] == kwargs['id']:
                        mem.update(m)
                        break

    def delete_member(self, id, pool_id):
        for p in self.wf_apply_params['parameters']['pools']:
            if p['id'] == pool_id:
                for mem in p['members']:
                    if mem['id'] == id:
                        index = p['members'].index(mem)
                        del p['members'][index]
                        break

    def delete_pool(self, id):
        for p in self.wf_apply_params['parameters']['pools']:
            if p['id'] == id:
                index = self.wf_apply_params['parameters']['pools'].index(p)
                del self.wf_apply_params['parameters']['pools'][index]
                break
        for l in self.wf_apply_params['parameters']['listeners']:
            if l['default_pool']['id'] == id:
                index = self.wf_apply_params['parameters']['listeners']\
                    .index(l)
                del self.wf_apply_params['parameters']['listeners'][index]
                break

    def add_network_to_service(self, subnet):
        self.wf_srv_params['primary']['network']['portgroups'].append(
            subnet['subnet']['network_id'])
        self.wf_create_params['parameters']['service_params'] =\
            self.wf_srv_params

    def set_two_leg_mode(self, two_leg_mode):
        self.wf_create_params['parameters']['twoleg_enabled'] = two_leg_mode

    def compare_create_call(self):
        create_call = self.driver_rest_call_mock.mock_calls[2][1][2]
        self.assertEqual(create_call, self.wf_create_params)

    def compare_apply_call(self):
        def _sort_lists_by_id(o):
            if isinstance(o, list):
                o = sorted(o, key=lambda x: x['id'])
                for i in o:
                    i = _sort_lists_by_id(i)
            elif isinstance(o, dict):
                for k in o.keys():
                    o[k] = _sort_lists_by_id(o[k])
            return o

        apply_call_i = len(self.driver_rest_call_mock.mock_calls) - 2
        apply_call = self.driver_rest_call_mock.mock_calls[apply_call_i][1][2]
        apply_call = _sort_lists_by_id(apply_call)
        self.wf_apply_params = _sort_lists_by_id(self.wf_apply_params)
        self.assertEqual(apply_call, self.wf_apply_params)

    def setUp(self):
        super(TestLBaaSDriver, self).setUp()

        templates_to_return = [{'name': self.driver.workflow_template_name}]
        for t in self.driver.child_workflow_template_names:
            templates_to_return.append({'name': t})
        rest_call_function_mock.__dict__.update(
            {'RESPOND_WITH_ERROR': False, 'WORKFLOW_MISSING': True,
             'WORKFLOW_TEMPLATE_MISSING': True,
             'RESPOND_WITH_SERVER_DOWN': 200,
             'WF_TEMPLATES_TO_RETURN': templates_to_return})

        self.operation_completer_start_mock = mock.Mock(
            return_value=None)
        self.operation_completer_join_mock = mock.Mock(
            return_value=None)
        self.driver_rest_call_mock = mock.Mock(
            side_effect=rest_call_function_mock)
        self.flip_servers_mock = mock.Mock(
            return_value=None)
        self.recover_mock = mock.Mock(
            side_effect=_recover_function_mock)

        self.driver.completion_handler.start = (
            self.operation_completer_start_mock)
        self.driver.completion_handler.join = (
            self.operation_completer_join_mock)
        self.driver.rest_client.call = self.driver_rest_call_mock
        self.driver.rest_client._call = self.driver_rest_call_mock
        self.driver.completion_handler.rest_client.call = (
            self.driver_rest_call_mock)

        self.driver.queue = QueueMock(
            self.driver.completion_handler.handle_operation_completion)

        self.wf_srv_params = copy.deepcopy(WF_SRV_PARAMS)
        self.wf_create_params = copy.deepcopy(WF_CREATE_PARAMS)
        self.wf_create_params['tenants'] = [self._tenant_id]
        self.wf_apply_params = copy.deepcopy(WF_APPLY_PARAMS)

        self.addCleanup(self.driver.completion_handler.join)

    def test_verify_workflow_templates(self):
        templates_to_return = []
        for t in self.driver.child_workflow_template_names:
            templates_to_return.append({'name': t})
        rest_call_function_mock.__dict__.update(
            {'WF_TEMPLATES_TO_RETURN': templates_to_return})
        message = r_exc.WorkflowTemplateMissing.message % \
            {'workflow_template': self.driver.workflow_template_name}
        try:
            self.driver._verify_workflow_templates()
        except r_exc.WorkflowTemplateMissing as e:
            self.assertEqual(message, e.msg)

        templates_to_return.append(
            {'name': self.driver.workflow_template_name})
        rest_call_function_mock.__dict__.update(
            {'WF_TEMPLATES_TO_RETURN': templates_to_return})
        try:
            self.driver._verify_workflow_templates()
            self.assertTrue(True)
        except r_exc.WorkflowTemplateMissing as e:
            self.assertTrue(False)

    def test_wf_created_on_first_member_creation(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub:
            with self.loadbalancer(subnet=vip_sub) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as l:
                    listener_id = l['listener']['id']
                    with self.pool(
                        protocol=lb_con.PROTOCOL_HTTP,
                        listener_id=listener_id) as p:
                        pool_id = p['pool']['id']
                        self.driver_rest_call_mock.reset_mock()
                        with self.member(
                            pool_id=pool_id,
                            subnet=vip_sub, address='10.0.1.10'):

                            self.compare_create_call()
                            self.compare_apply_call()

    def test_wf_deleted_on_lb_deletion(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub:
            with self.loadbalancer(subnet=vip_sub) as lb:
                get_calls = [
                    mock.call('GET', u'/api/workflow/LB_' +
                        lb['loadbalancer']['id'], None, None)]
                with self.listener(
                    loadbalancer_id=lb['loadbalancer']['id']) as listener:
                    with self.pool(
                        protocol=lb_con.PROTOCOL_HTTP,
                        listener_id=listener['listener']['id']) as pool:
                        with self.member(pool_id=pool['pool']['id'],
                                         subnet=vip_sub, address='10.0.1.10'):
                            self.driver_rest_call_mock.reset_mock()
                            rest_call_function_mock.__dict__.update(
                                {'WORKFLOW_MISSING': False})

                        self.driver_rest_call_mock.assert_has_calls(get_calls)
                        self.driver_rest_call_mock.reset_mock()
                    self.driver_rest_call_mock.assert_has_calls(get_calls)
                    self.driver_rest_call_mock.reset_mock()
                self.driver_rest_call_mock.assert_has_calls(get_calls)
                self.driver_rest_call_mock.reset_mock()
            self.driver_rest_call_mock.assert_any_call(
                'DELETE', u'/api/workflow/LB_' + lb['loadbalancer']['id'],
                None, None)

    def test_lb_crud(self):
        with self.subnet(cidr='10.0.0.0/24') as s:
            with self.loadbalancer(subnet=s, no_delete=True) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as l:
                    with self.pool(
                        protocol=lb_con.PROTOCOL_HTTP,
                        listener_id=l['listener']['id']) as p:

                        self.plugin_instance.update_loadbalancer(
                            context.get_admin_context(),
                            lb_id, {'loadbalancer': lb})
                        lb_db = self.plugin_instance.db.get_loadbalancer(
                            context.get_admin_context(),
                            lb_id)
                        self.driver.load_balancer.refresh(
                            context.get_admin_context(), lb_db)

                        with self.member(
                            no_delete=True, pool_id=p['pool']['id'],
                            subnet=s, address='10.0.1.10'):

                            self.compare_apply_call()

                            self.driver_rest_call_mock.reset_mock()
                            rest_call_function_mock.__dict__.update(
                                {'WORKFLOW_MISSING': False})

                            self.plugin_instance.update_loadbalancer(
                                context.get_admin_context(),
                                lb_id, {'loadbalancer': lb})
                            self.compare_apply_call()

                            self.driver_rest_call_mock.reset_mock()
                            lb_db = self.plugin_instance.db.get_loadbalancer(
                                context.get_admin_context(), lb_id)
                            self.driver.load_balancer.refresh(
                                context.get_admin_context(), lb_db)
                            self.compare_apply_call()

                self.plugin_instance.delete_loadbalancer(
                    context.get_admin_context(), lb_id)
                self.driver_rest_call_mock.assert_any_call(
                    'DELETE', '/api/workflow/LB_' + lb_id,
                    None, None)
                self.assertRaises(loadbalancerv2.EntityNotFound,
                                  self.plugin_instance.get_loadbalancer,
                                  context.get_admin_context(), lb_id)

    def test_lb_stats(self):
        with self.subnet(cidr='10.0.0.0/24') as s:
            with self.loadbalancer(subnet=s) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as l:
                    with self.pool(
                        protocol=lb_con.PROTOCOL_HTTP,
                        listener_id=l['listener']['id']) as p:
                        with self.member(
                            no_delete=True, pool_id=p['pool']['id'],
                            subnet=s, address='10.0.1.10'):

                            rest_call_function_mock.__dict__.update(
                                {'WORKFLOW_MISSING': False})

                            stats = self.plugin_instance.stats(
                                context.get_admin_context(), lb_id,)
                            self.assertEqual({'stats': {'bytes_in': 100,
                                'total_connections': 2,
                                'active_connections': 1, 'bytes_out': 200}},
                                stats)

    def test_member_crud(self):
        with self.subnet(cidr='10.0.0.0/24') as s:
            with self.loadbalancer(subnet=s) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as l:
                    listener_id = l['listener']['id']
                    with self.pool(
                        protocol=lb_con.PROTOCOL_HTTP,
                        listener_id=listener_id) as p:
                        pool_id = p['pool']['id']
                        with self.member(
                            no_delete=True, address='10.0.1.10',
                            pool_id=pool_id, subnet=s) as m1:
                            member1_id = m1['member']['id']

                            self.driver_rest_call_mock.reset_mock()
                            rest_call_function_mock.__dict__.update(
                                {'WORKFLOW_MISSING': False})

                            with self.member(
                                no_delete=True, pool_id=pool_id,
                                subnet=s, address='10.0.1.20') as m2:
                                member2_id = m2['member']['id']
                                self.compare_apply_call()

                                self.driver_rest_call_mock.reset_mock()
                                m = self.plugin_instance.db.get_pool_member(
                                    context.get_admin_context(),
                                    m1['member']['id']).to_dict(pool=False)

                                m['weight'] = 2
                                self.plugin_instance.update_pool_member(
                                    context.get_admin_context(),
                                    m1['member']['id'], p['pool']['id'],
                                    {'member': m})
                                self.update_member(pool_id, id=member1_id,
                                                   weight=2)
                                self.compare_apply_call()

                                self.driver_rest_call_mock.reset_mock()

                                self.plugin_instance.delete_pool_member(
                                    context.get_admin_context(),
                                    member2_id, pool_id)
                                self.delete_member(member2_id, pool_id)
                                self.compare_apply_call()

                                lb = self.plugin_instance.db.get_loadbalancer(
                                    context.get_admin_context(),
                                    lb_id).to_dict(listener=False)
                                self.assertEqual('ACTIVE',
                                             lb['provisioning_status'])

    def test_build_objects_with_tls(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub:
            with self.loadbalancer(subnet=vip_sub) as lb:
                lb_id = lb['loadbalancer']['id']
                with mock.patch('neutron_lbaas.services.loadbalancer.plugin.'
                                'cert_parser',
                                autospec=True) as cert_parser_mock, \
                        mock.patch('neutron_lbaas.services.loadbalancer.'
                                   'plugin.CERT_MANAGER_PLUGIN.CertManager',
                                   autospec=True) as cert_manager_mock:
                    cert_mock = mock.Mock(spec=cert_manager.Cert)
                    cert_mock.get_certificate.return_value = 'certificate'
                    cert_mock.get_intermediates.return_value = 'intermediates'
                    cert_mock.get_private_key.return_value = 'private_key'
                    cert_mock.get_private_key_passphrase.return_value = \
                        'private_key_passphrase'
                    cert_manager_mock().get_cert.return_value = cert_mock
                    cert_parser_mock.validate_cert.return_value = True

                    with self.listener(
                        protocol=lb_con.PROTOCOL_TERMINATED_HTTPS,
                        loadbalancer_id=lb_id,
                        default_tls_container_ref='def1',
                        sni_container_refs=['sni1', 'sni2']) as listener:
                        with self.pool(
                            protocol=lb_con.PROTOCOL_HTTP,
                            listener_id=listener['listener']['id']) as pool:
                            with self.member(pool_id=pool['pool']['id'],
                                             subnet=vip_sub,
                                             address='10.0.1.10'):
                                self.compare_apply_call()

    def test_build_objects_with_l7(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub:
            with self.loadbalancer(subnet=vip_sub) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(
                    protocol=lb_con.PROTOCOL_HTTP,
                    loadbalancer_id=lb_id) as listener:
                    listener_id = listener['listener']['id']
                    with self.pool(protocol=lb_con.PROTOCOL_HTTP,
                                   listener_id=listener_id) as def_pool, \
                            self.pool(protocol=lb_con.PROTOCOL_HTTP,
                                      loadbalancer_id=lb_id) as pol_pool:
                        def_pool_id = def_pool['pool']['id']
                        pol_pool_id = pol_pool['pool']['id']
                        with self.l7policy(
                            listener_id,
                            action=lb_con.L7_POLICY_ACTION_REDIRECT_TO_POOL,
                            redirect_pool_id=pol_pool_id) as policy:
                            policy_id = policy['l7policy']['id']

                            self.driver_rest_call_mock.reset_mock()
                            with self.l7policy_rule(l7policy_id=policy_id,
                                                    key=u'key1',
                                                    value=u'val1'), \
                                    self.l7policy_rule(l7policy_id=policy_id,
                                                       key=u'key2',
                                                       value=u'val2'), \
                                    self.member(pool_id=def_pool_id,
                                                subnet=vip_sub,
                                                address=u'10.0.1.10'):
                                self.driver_rest_call_mock.reset_mock()
                                rest_call_function_mock.__dict__.update(
                                    {'WORKFLOW_MISSING': False})

                                with self.member(
                                    pool_id=pol_pool_id,
                                    subnet=vip_sub,
                                    address=u'10.0.1.20'):

                                    self.compare_apply_call()

    def test_build_objects_graph_lb_pool(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub:
            with self.loadbalancer(subnet=vip_sub) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as listener:
                    listener_id = listener['listener']['id']
                    with self.pool(
                        protocol=lb_con.PROTOCOL_HTTP,
                        listener_id=listener_id) as pool:
                        with self.member(pool_id=pool['pool']['id'],
                                         subnet=vip_sub,
                                         address='10.0.1.10'), \
                                self.member(pool_id=pool['pool']['id'],
                                            subnet=vip_sub,
                                            address='10.0.1.20'):
                            self.driver_rest_call_mock.reset_mock()
                            rest_call_function_mock.__dict__.update(
                                {'WORKFLOW_MISSING': False})

                            with self.pool(
                                protocol=lb_con.PROTOCOL_HTTP,
                                loadbalancer_id=lb_id):
                                self.compare_apply_call()

    def test_build_objects_graph_one_leg(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub:
            with self.loadbalancer(subnet=vip_sub) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as listener:
                    listener_id = listener['listener']['id']
                    with self.pool(
                        protocol='HTTP',
                        listener_id=listener_id) as pool:
                        with self.member(pool_id=pool['pool']['id'],
                                         subnet=vip_sub,
                                         address='10.0.1.10'), \
                                self.member(pool_id=pool['pool']['id'],
                                            subnet=vip_sub,
                                            address='10.0.1.20'):
                            self.compare_apply_call()

    def test_build_objects_graph_two_legs_full(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub, \
                self.subnet(cidr='20.0.0.0/24') as member_sub1, \
                self.subnet(cidr='30.0.0.0/24'):
            with self.loadbalancer(subnet=vip_sub) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as listener:
                    with self.pool(
                        protocol='HTTP',
                        listener_id=listener['listener']['id'],
                        session_persistence={
                            'type': "APP_COOKIE",
                            'cookie_name': 'sessionId'}) as pool:
                        with self.healthmonitor(
                            type='HTTP', pool_id=pool['pool']['id']):

                            self.driver_rest_call_mock.reset_mock()

                            with self.member(
                                pool_id=pool['pool']['id'],
                                subnet=member_sub1,
                                address='20.0.1.10') as member:

                                    self.update_member(
                                        pool['pool']['id'],
                                        id=member['member']['id'],
                                        address='20.0.1.10',
                                        subnet='20.0.1.10', gw='20.0.0.1')
                                    self.set_two_leg_mode(True)
                                    self.add_network_to_service(member_sub1)
                                    self.wf_apply_params['parameters'][
                                        'pip_address'] =\
                                        self.driver_rest_call_mock.mock_calls[
                                            len(self.driver_rest_call_mock.
                                            mock_calls) - 2][1][2][
                                                'parameters']['pip_address']

                                    self.compare_create_call()
                                    self.compare_apply_call()

    def test_pool_deletion_for_listener(self):
        with self.subnet(cidr='10.0.0.0/24') as vip_sub:
            with self.loadbalancer(subnet=vip_sub) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as listener:
                    with self.pool(
                        protocol='HTTP',
                        listener_id=listener['listener']['id'],
                        no_delete=True) as p:

                        with self.member(
                            no_delete=True,
                            pool_id=p['pool']['id'],
                            subnet=vip_sub, address='10.0.1.10'):

                            self.driver_rest_call_mock.reset_mock()
                            rest_call_function_mock.__dict__.update(
                                {'WORKFLOW_MISSING': False})

                            self.plugin_instance.delete_pool(
                                context.get_admin_context(), p['pool']['id'])
                            self.delete_pool(p['pool']['id'])

                            self.compare_apply_call()


class TestLBaaSDriverDebugOptions(TestLBaaSDriverBase):
    def setUp(self):
        cfg.CONF.set_override('configure_l3', False,
                              group='radwarev2_debug')
        cfg.CONF.set_override('configure_l4', False,
                              group='radwarev2_debug')
        super(TestLBaaSDriverDebugOptions, self).setUp()

        templates_to_return = [{'name': self.driver.workflow_template_name}]
        for t in self.driver.child_workflow_template_names:
            templates_to_return.append({'name': t})
        rest_call_function_mock.__dict__.update(
            {'RESPOND_WITH_ERROR': False, 'WORKFLOW_MISSING': True,
             'WORKFLOW_TEMPLATE_MISSING': True,
             'RESPOND_WITH_SERVER_DOWN': 200,
             'WF_TEMPLATES_TO_RETURN': templates_to_return})

        self.operation_completer_start_mock = mock.Mock(
            return_value=None)
        self.operation_completer_join_mock = mock.Mock(
            return_value=None)
        self.driver_rest_call_mock = mock.Mock(
            side_effect=rest_call_function_mock)
        self.flip_servers_mock = mock.Mock(
            return_value=None)
        self.recover_mock = mock.Mock(
            side_effect=_recover_function_mock)

        self.driver.completion_handler.start = (
            self.operation_completer_start_mock)
        self.driver.completion_handler.join = (
            self.operation_completer_join_mock)
        self.driver.rest_client.call = self.driver_rest_call_mock
        self.driver.rest_client._call = self.driver_rest_call_mock
        self.driver.completion_handler.rest_client.call = (
            self.driver_rest_call_mock)

        self.driver.queue = QueueMock(
            self.driver.completion_handler.handle_operation_completion)

    def test_debug_options(self):
        with self.subnet(cidr='10.0.0.0/24') as s:
            with self.loadbalancer(subnet=s) as lb:
                lb_id = lb['loadbalancer']['id']
                with self.listener(loadbalancer_id=lb_id) as l:
                    with self.pool(
                        protocol='HTTP',
                        listener_id=l['listener']['id']) as p:
                        with self.member(
                            pool_id=p['pool']['id'],
                            subnet=s, address='10.0.1.10'):
                            wf_srv_params = copy.deepcopy(WF_SRV_PARAMS)
                            wf_params = copy.deepcopy(WF_CREATE_PARAMS)

                            wf_srv_params['name'] = 'srv_' + (
                                s['subnet']['network_id'])
                            wf_srv_params['tenantId'] = self._tenant_id
                            wf_srv_params['primary']['network'][
                                'portgroups'] = [s['subnet'][
                                     'network_id']]
                            wf_params['tenants'] = [self._tenant_id]
                            wf_params['parameters']['service_params'] = (
                                wf_srv_params)
                            wf_params['parameters']['configure_l3'] = False
                            wf_params['parameters']['configure_l4'] = False
                            calls = [
                                mock.call('GET', '/api/workflow/LB_' + lb_id,
                                          None, None),
                                mock.call(
                                    'POST',
                                    '/api/workflowTemplate/' +
                                    'os_lb_v2?name=LB_' + lb_id,
                                    wf_params,
                                    v2_driver.TEMPLATE_HEADER)
                            ]
                            self.driver_rest_call_mock.assert_has_calls(
                                calls, any_order=True)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest import config

from neutron_lbaas.tests.tempest.v2.scenario import base


CONF = config.CONF


class TestSessionPersistence(base.BaseTestCase):

    @utils.services('compute', 'network')
    def test_session_persistence(self):
        """This test checks checks load balancing with session persistence.

        The following is the scenario outline:
        1. Boot two instances.
        2. SSH to the instance and start two servers.
        3. Create a pool with SOURCE_IP session persistence type.
        4. Create a load balancer with two members and with ROUND_ROBIN
           algorithm.
        5. Send 10 requests to the floating ip, associated with the VIP,
           and make sure all the requests from the same ip
           are processed by the same member of the pool.
        6. Change session persistence type of the pool to HTTP_COOKIE.
        7. Check that this session persistence type also forces all
           the requests containing the same cookie to hit the same
           member of the pool.
        8. Change session persistence type of the pool to APP_COOKIE.
        9. Perform the same check.
        10. Turn session persistence off and check that the requests
            are again distributed according to the ROUND_ROBIN algorithm.
        """
        self._create_server('server1')
        self._start_servers()
        session_persistence_types = CONF.lbaas.session_persistence_types
        if "SOURCE_IP" in session_persistence_types:
            self._create_load_balancer(persistence_type="SOURCE_IP")
            self._check_source_ip_persistence()
        if "HTTP_COOKIE" in session_persistence_types:
            self._update_pool_session_persistence("HTTP_COOKIE")
            self._check_cookie_session_persistence()
        if "APP_COOKIE" in session_persistence_types:
            self._update_pool_session_persistence("APP_COOKIE",
                                                  cookie_name="JSESSIONID")
            self._check_cookie_session_persistence()
        self._update_pool_session_persistence()
        self._check_load_balancing()



# Copyright 2015 Rackspace.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""agentv2

Revision ID: 364f9b6064f0
Revises: 4b6d8d5310b8
Create Date: 2015-02-05 10:17:13.229358

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '364f9b6064f0'
down_revision = '4b6d8d5310b8'


def upgrade():
    op.create_table(
        'lbaas_loadbalanceragentbindings',
        sa.Column('loadbalancer_id', sa.String(length=36), nullable=False),
        sa.Column('agent_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['loadbalancer_id'],
                                ['lbaas_loadbalancers.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['agent_id'], ['agents.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('loadbalancer_id'))



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base
CONF = config.CONF


class MemberTestJSON(base.BaseAdminTestCase):
    """
    Test the member creation operation in admin scope in
    Neutron-LBaaS API using the REST client for members:

    """

    @classmethod
    def resource_setup(cls):
        super(MemberTestJSON, cls).resource_setup()
        if not utils.is_extension_enabled("lbaasv2", "network"):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.tenant_id = cls.subnet.get('tenant_id')
        cls.subnet_id = cls.subnet.get('id')
        cls.load_balancer = cls._create_active_load_balancer(
            tenant_id=cls.tenant_id,
            vip_subnet_id=cls.subnet.get('id'))
        cls.load_balancer_id = cls.load_balancer.get("id")
        cls._wait_for_load_balancer_status(cls.load_balancer_id)
        cls.listener = cls._create_listener(
            loadbalancer_id=cls.load_balancer.get('id'),
            protocol=cls.listener_protocol, protocol_port=80)
        cls.listener_id = cls.listener.get('id')
        cls.pool = cls._create_pool(protocol=cls.pool_protocol,
                                    tenant_id=cls.tenant_id,
                                    lb_algorithm='ROUND_ROBIN',
                                    listener_id=cls.listener_id)
        cls.pool_id = cls.pool.get('id')

    @classmethod
    def resource_cleanup(cls):
        super(MemberTestJSON, cls).resource_cleanup()

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='smoke')
    def test_create_member_invalid_tenant_id(self):
        """Test create member with invalid tenant_id"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['tenant_id'] = "$232!$pw"
        member = self._create_member(self.pool_id, **member_opts)
        self.addCleanup(self._delete_member, self.pool_id, member['id'])
        self.assertEqual(member['subnet_id'], self.subnet_id)
        self.assertEqual(member['tenant_id'], "$232!$pw")

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_member_empty_tenant_id(self):
        """Test create member with an empty tenant_id should fail"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['tenant_id'] = ""
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)



# Copyright 2015, Banashankar Veerad, Copyright IBM Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from neutron_lib import context
from oslo_config import cfg

from neutron_lbaas.common import exceptions
from neutron_lbaas.drivers.octavia import driver
from neutron_lbaas.services.loadbalancer import constants
from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2


class ManagerTest(object):
    def __init__(self, parent, manager, mocked_req):
        self.parent = parent
        self.context = parent.context
        self.driver = parent.driver
        self.manager = manager
        self.mocked_req = mocked_req

    def create(self, model, url, args):
        self.manager.create(self.context, model)
        self.mocked_req.post.assert_called_with(url, args)

    def update(self, old_model, model, url, args):
        self.manager.update(self.context, old_model, model)
        self.mocked_req.put.assert_called_with(url, args)

    def delete(self, model, url):
        self.manager.delete(self.context, model)
        self.mocked_req.delete.assert_called_with(url)

    def delete_cascade(self, model, url):
        self.manager.delete_cascade(self.context, model)
        self.mocked_req.delete.assert_called_with(url)

    # TODO(Banashankar) : Complete refresh function. Need more info.
    def refresh(self):
        pass

    # TODO(Banashankar): Complete stats function. Need more info.
    def stats(self):
        pass


class BaseOctaviaDriverTest(test_db_loadbalancerv2.LbaasPluginDbTestCase):

    # Copied it from Brocade's test code :/
    def _create_fake_models(self, children=True, graph=False):
        # This id is used for all the entities.
        id = 'test_id'
        lb = data_models.LoadBalancer(id=id)
        if not children:
            return lb
        sni_container = data_models.SNI(listener_id=id)
        listener = data_models.Listener(id=id, loadbalancer=lb,
                                        sni_containers=[sni_container])
        pool = data_models.Pool(id=id, loadbalancer=lb)
        member = data_models.Member(id=id, pool=pool)
        hm = data_models.HealthMonitor(id=id, pool=pool)
        sp = data_models.SessionPersistence(pool_id=pool.id, pool=pool)
        l7policy = data_models.L7Policy(
            id=id, listener=listener, listener_id=listener.id,
            action=constants.L7_POLICY_ACTION_REDIRECT_TO_POOL)
        l7rule = data_models.L7Rule(
            id=id, policy=l7policy, type=constants.L7_RULE_TYPE_PATH,
            compare_type=constants.L7_RULE_COMPARE_TYPE_STARTS_WITH,
            value='/api')
        lb.listeners = [listener]
        lb.pools = [pool]
        if graph:
            r_pool = data_models.Pool(id=id, loadbalancer=lb)
            r_member = data_models.Member(id=id, pool=r_pool)
            r_pool.members = [r_member]
            l7policy.redirect_pool = r_pool
            l7policy.redirect_pool_id = r_pool.id
            lb.pools.append(r_pool)
        else:
            l7policy.redirect_pool = pool
            l7policy.redirect_pool_id = pool.id
        listener.default_pool = pool
        listener.l7_policies = [l7policy]
        l7policy.rules = [l7rule]
        pool.members = [member]
        pool.session_persistence = sp
        pool.healthmonitor = hm

        return lb

    def setUp(self):
        super(BaseOctaviaDriverTest, self).setUp()
        self.context = context.get_admin_context()
        self.plugin = mock.Mock()
        self.driver = driver.OctaviaDriver(self.plugin)
        # mock of rest call.
        self.driver.req = mock.Mock()
        self.lb = self._create_fake_models()


class TestOctaviaDriver(BaseOctaviaDriverTest):

    def test_allocates_vip(self):
        self.addCleanup(cfg.CONF.clear_override,
                        'allocates_vip', group='octavia')
        cfg.CONF.set_override('allocates_vip', True, group='octavia')
        test_driver = driver.OctaviaDriver(self.plugin)
        self.assertTrue(test_driver.load_balancer.allocates_vip)

    def test_create_load_balancer_graph(self):
        m = ManagerTest(self, self.driver.load_balancer,
                        self.driver.req)
        lb = self._create_fake_models(children=True, graph=True)
        listener = lb.listeners[0]
        sni_container = listener.sni_containers[0]
        policy = listener.l7_policies[0]
        r_pool = policy.redirect_pool
        r_member = r_pool.members[0]
        rule = policy.rules[0]
        pool = listener.default_pool
        member = pool.members[0]
        sp = pool.session_persistence
        hm = pool.healthmonitor

        lb_url = '/v1/loadbalancers'

        args = {
            "id": lb.id,
            "name": lb.name,
            "enabled": lb.admin_state_up,
            "vip": {
                "subnet_id": lb.vip_subnet_id,
                "port_id": lb.vip_port_id,
                "ip_address": lb.vip_address
            },
            "listeners": [{
                "id": listener.id,
                "protocol": listener.protocol,
                "enabled": listener.admin_state_up,
                "sni_containers": [sni_container.tls_container_id],
                "tls_certificate_id": listener.default_tls_container_id,
                "l7policies": [{
                    "id": policy.id,
                    "name": policy.name,
                    "redirect_pool": {
                        "id": r_pool.id,
                        "lb_algorithm": r_pool.lb_algorithm,
                        "protocol": r_pool.protocol,
                        "name": r_pool.name,
                        "enabled": r_pool.admin_state_up,
                        "session_persistence": r_pool.session_persistence,
                        "members": [{
                            "project_id": r_member.tenant_id,
                            "weight": r_member.weight,
                            "subnet_id": r_member.subnet_id,
                            "ip_address": r_member.address,
                            "protocol_port": r_member.protocol_port,
                            "enabled": r_member.admin_state_up,
                            "id": r_member.id
                        }],
                        "project_id": r_pool.tenant_id,
                        "description": r_pool.description
                    },
                    "l7rules": [{
                        "id": rule.id,
                        "type": rule.type,
                        "compare_type": rule.compare_type,
                        "key": rule.key,
                        "value": rule.value,
                        "invert": rule.invert
                    }],
                    "enabled": policy.admin_state_up,
                    "action": policy.action,
                    "position": policy.position,
                    "description": policy.description
                }],
                "name": listener.name,
                "description": listener.description,
                "default_pool": {
                    "id": pool.id,
                    "lb_algorithm": pool.lb_algorithm,
                    "protocol": pool.protocol,
                    "name": pool.name,
                    "enabled": pool.admin_state_up,
                    "session_persistence": {
                        "cookie_name": sp.cookie_name,
                        "type": sp.type
                    },
                    "members": [{
                        "project_id": member.tenant_id,
                        "weight": member.weight,
                        "subnet_id": member.subnet_id,
                        "ip_address": member.address,
                        "protocol_port": member.protocol_port,
                        "enabled": member.admin_state_up,
                        "id": member.id
                    }],
                    "health_monitor": {
                        'type': hm.type,
                        'delay': hm.delay,
                        'timeout': hm.timeout,
                        'rise_threshold': hm.max_retries,
                        'fall_threshold': hm.max_retries,
                        'http_method': hm.http_method,
                        'url_path': hm.url_path,
                        'expected_codes': hm.expected_codes,
                        'enabled': hm.admin_state_up,
                        'project_id': hm.tenant_id
                    },
                    "project_id": pool.tenant_id,
                    "description": pool.description
                },
                "connection_limit": listener.connection_limit,
                "protocol_port": listener.protocol_port,
                "project_id": listener.tenant_id
            }],
            "project_id": lb.tenant_id,
            "description": lb.description
        }

        m.create(lb, lb_url, args)

    def test_load_balancer_ops(self):
        m = ManagerTest(self, self.driver.load_balancer,
                        self.driver.req)

        lb = self._create_fake_models(children=False)

        # urls for assert test.
        lb_url = '/v1/loadbalancers'
        lb_url_id = '/v1/loadbalancers/' + lb.id

        # Create LB test
        # args for create assert.
        args = {
            'id': lb.id,
            'name': lb.name,
            'description': lb.description,
            'enabled': lb.admin_state_up,
            'project_id': lb.tenant_id,
            'vip': {
                'subnet_id': lb.vip_subnet_id,
                'ip_address': lb.vip_address,
                'port_id': lb.vip_port_id,
            }
        }
        m.create(lb, lb_url, args)

        # Update LB test
        # args for update assert.
        args = {
            'name': lb.name,
            'description': lb.description,
            'enabled': lb.admin_state_up,
        }
        m.update(lb, lb, lb_url_id, args)

        # delete LB test
        m.delete_cascade(lb, lb_url_id + '/delete_cascade')

        # TODO(Banashankar) : refresh n stats fucntions are not yet done.
        #m.refresh()
        #m.stats()

    def test_listener_ops(self):
        m = ManagerTest(self, self.driver.listener,
                        self.driver.req)

        listener = self.lb.listeners[0]

        # urls for assert test.
        list_url = '/v1/loadbalancers/%s/listeners' % listener.loadbalancer.id
        list_url_id = list_url + '/%s' % (listener.id)

        # Create Listener test.
        # args for create and update assert.
        sni_containers = [sni.tls_container_id
                          for sni in listener.sni_containers]
        args = {
            'id': listener.id,
            'name': listener.name,
            'description': listener.description,
            'enabled': listener.admin_state_up,
            'protocol': listener.protocol,
            'protocol_port': listener.protocol_port,
            'connection_limit': listener.connection_limit,
            'tls_certificate_id': listener.default_tls_container_id,
            'sni_containers': sni_containers,
            'default_pool_id': listener.default_pool_id,
            'project_id': listener.tenant_id
        }
        m.create(listener, list_url, args)

        # Update listener test.
        del args['id']
        del args['project_id']
        m.update(listener, listener, list_url_id, args)

        # Delete listener.
        m.delete(listener, list_url_id)

    def test_pool_ops(self):
        m = ManagerTest(self, self.driver.pool,
                        self.driver.req)

        pool = self.lb.listeners[0].default_pool

        # urls for assert test.
        pool_url = '/v1/loadbalancers/%s/pools' % (
            pool.loadbalancer.id)
        pool_url_id = pool_url + "/%s" % pool.id

        # Test create pool.
        # args for create and update assert.
        args = {
            'id': pool.id,
            'name': pool.name,
            'description': pool.description,
            'enabled': pool.admin_state_up,
            'protocol': pool.protocol,
            'lb_algorithm': pool.lb_algorithm,
            'project_id': pool.tenant_id
        }
        if pool.session_persistence:
            args['session_persistence'] = {
                'type': pool.session_persistence.type,
                'cookie_name': pool.session_persistence.cookie_name,
            }
        else:
            args['session_persistence'] = None
        m.create(pool, pool_url, args)

        # Test update pool.
        del args['id']
        del args['project_id']
        m.update(pool, pool, pool_url_id, args)

        # Test pool delete.
        m.delete(pool, pool_url_id)

    def test_member_ops(self):
        m = ManagerTest(self, self.driver.member,
                        self.driver.req)

        member = self.lb.listeners[0].default_pool.members[0]

        # urls for assert.
        mem_url = '/v1/loadbalancers/%s/pools/%s/members' % (
            member.pool.loadbalancer.id,
            member.pool.id)
        mem_url_id = mem_url + "/%s" % member.id

        # Test Create member.
        # args for create assert.
        args = {
            'id': member.id,
            'enabled': member.admin_state_up,
            'ip_address': member.address,
            'protocol_port': member.protocol_port,
            'weight': member.weight,
            'subnet_id': member.subnet_id,
            'project_id': member.tenant_id
        }
        m.create(member, mem_url, args)

        # Test member update.
        # args for update assert.
        args = {
            'enabled': member.admin_state_up,
            'protocol_port': member.protocol_port,
            'weight': member.weight,
        }
        m.update(member, member, mem_url_id, args)

        # Test member delete.
        m.delete(member, mem_url_id)

    def test_health_monitor_ops(self):
        m = ManagerTest(self, self.driver.health_monitor,
                        self.driver.req)

        hm = self.lb.listeners[0].default_pool.healthmonitor

        # urls for assert.
        hm_url = '/v1/loadbalancers/%s/pools/%s/healthmonitor' % (
            hm.pool.loadbalancer.id,
            hm.pool.id)

        # Test HM create.
        # args for create and update assert.
        args = {
            'type': hm.type,
            'delay': hm.delay,
            'timeout': hm.timeout,
            'rise_threshold': hm.max_retries,
            'fall_threshold': hm.max_retries_down,
            'http_method': hm.http_method,
            'url_path': hm.url_path,
            'expected_codes': hm.expected_codes,
            'enabled': hm.admin_state_up,
            'project_id': hm.tenant_id
        }
        m.create(hm, hm_url, args)

        # Test HM update
        del args['project_id']
        m.update(hm, hm, hm_url, args)

        # Test HM delete
        m.delete(hm, hm_url)

    def test_l7_policy_ops_reject(self):
        m = ManagerTest(self, self.driver.l7policy,
                        self.driver.req)

        l7p = copy.deepcopy(self.lb.listeners[0].l7_policies[0])
        l7p.action = constants.L7_POLICY_ACTION_REJECT

        # urls for assert.
        l7p_url = '/v1/loadbalancers/%s/listeners/%s/l7policies' % (
            l7p.listener.loadbalancer.id,
            l7p.listener.id)
        l7p_url_id = l7p_url + "/%s" % l7p.id

        # Test L7Policy create.
        # args for create and update assert.
        args = {
            'id': l7p.id,
            'name': l7p.name,
            'description': l7p.description,
            'action': constants.L7_POLICY_ACTION_REJECT,
            'position': l7p.position,
            'enabled': l7p.admin_state_up
        }
        m.create(l7p, l7p_url, args)

        # Test L7Policy update
        del args['id']
        m.update(l7p, l7p, l7p_url_id, args)

        # Test L7Policy delete
        m.delete(l7p, l7p_url_id)

    def test_l7_policy_ops_rdr_pool(self):
        m = ManagerTest(self, self.driver.l7policy,
                        self.driver.req)

        l7p = copy.deepcopy(self.lb.listeners[0].l7_policies[0])
        l7p.action = constants.L7_POLICY_ACTION_REDIRECT_TO_POOL

        # urls for assert.
        l7p_url = '/v1/loadbalancers/%s/listeners/%s/l7policies' % (
            l7p.listener.loadbalancer.id,
            l7p.listener.id)
        l7p_url_id = l7p_url + "/%s" % l7p.id

        # Test L7Policy create.
        # args for create and update assert.
        args = {
            'id': l7p.id,
            'name': l7p.name,
            'description': l7p.description,
            'action': constants.L7_POLICY_ACTION_REDIRECT_TO_POOL,
            'redirect_pool_id': l7p.redirect_pool_id,
            'position': l7p.position,
            'enabled': l7p.admin_state_up
        }
        m.create(l7p, l7p_url, args)

        # Test L7Policy update
        del args['id']
        m.update(l7p, l7p, l7p_url_id, args)

        # Test L7Policy delete
        m.delete(l7p, l7p_url_id)

    def test_l7_policy_ops_rdr_url(self):
        m = ManagerTest(self, self.driver.l7policy,
                        self.driver.req)

        l7p = copy.deepcopy(self.lb.listeners[0].l7_policies[0])
        l7p.action = constants.L7_POLICY_ACTION_REDIRECT_TO_URL

        # urls for assert.
        l7p_url = '/v1/loadbalancers/%s/listeners/%s/l7policies' % (
            l7p.listener.loadbalancer.id,
            l7p.listener.id)
        l7p_url_id = l7p_url + "/%s" % l7p.id

        # Test L7Policy create.
        # args for create and update assert.
        args = {
            'id': l7p.id,
            'name': l7p.name,
            'description': l7p.description,
            'action': constants.L7_POLICY_ACTION_REDIRECT_TO_URL,
            'redirect_url': l7p.redirect_url,
            'position': l7p.position,
            'enabled': l7p.admin_state_up
        }
        m.create(l7p, l7p_url, args)

        # Test L7Policy update
        del args['id']
        m.update(l7p, l7p, l7p_url_id, args)

        # Test L7Policy delete
        m.delete(l7p, l7p_url_id)

    def test_l7_rule_ops(self):
        m = ManagerTest(self, self.driver.l7rule,
                        self.driver.req)

        l7r = self.lb.listeners[0].l7_policies[0].rules[0]

        # urls for assert.
        l7r_url = '/v1/loadbalancers/%s/listeners/%s/l7policies/%s/l7rules' % (
            l7r.policy.listener.loadbalancer.id,
            l7r.policy.listener.id,
            l7r.policy.id)
        l7r_url_id = l7r_url + "/%s" % l7r.id

        # Test L7Rule create.
        # args for create and update assert.
        args = {
            'id': l7r.id,
            'type': l7r.type,
            'compare_type': l7r.compare_type,
            'key': l7r.key,
            'value': l7r.value,
            'invert': l7r.invert
        }
        m.create(l7r, l7r_url, args)

        # Test L7rule update
        del args['id']
        m.update(l7r, l7r, l7r_url_id, args)

        # Test L7Rule delete
        m.delete(l7r, l7r_url_id)


class TestThreadedDriver(BaseOctaviaDriverTest):

        def setUp(self):
            super(TestThreadedDriver, self).setUp()
            cfg.CONF.set_override('request_poll_interval', 1, group='octavia')
            cfg.CONF.set_override('request_poll_timeout', 5, group='octavia')
            self.driver.req.get = mock.MagicMock()
            self.succ_completion = mock.MagicMock()
            self.fail_completion = mock.MagicMock()
            self.context = mock.MagicMock()
            ctx_patcher = mock.patch('neutron_lib.context.get_admin_context',
                                     return_value=self.context)
            ctx_patcher.start()
            self.addCleanup(ctx_patcher.stop)
            self.driver.load_balancer.successful_completion = (
                self.succ_completion)
            self.driver.load_balancer.failed_completion = self.fail_completion

        def test_thread_op_goes_active(self):
            self.driver.req.get.side_effect = [
                {'provisioning_status': 'PENDING_CREATE'},
                {'provisioning_status': 'ACTIVE'}
            ]
            driver.thread_op(self.driver.load_balancer, self.lb)
            self.succ_completion.assert_called_once_with(self.context, self.lb,
                                                         delete=False)
            self.assertEqual(0, self.fail_completion.call_count)

        def test_thread_op_goes_deleted(self):
            self.driver.req.get.side_effect = [
                {'provisioning_status': 'PENDING_DELETE'},
                {'provisioning_status': 'DELETED'}
            ]
            driver.thread_op(self.driver.load_balancer, self.lb, delete=True)
            self.succ_completion.assert_called_once_with(self.context, self.lb,
                                                         delete=True)
            self.assertEqual(0, self.fail_completion.call_count)

        def test_thread_op_goes_error(self):
            self.driver.req.get.side_effect = [
                {'provisioning_status': 'PENDING_CREATE'},
                {'provisioning_status': 'ERROR'}
            ]
            driver.thread_op(self.driver.load_balancer, self.lb)
            self.fail_completion.assert_called_once_with(self.context, self.lb)
            self.assertEqual(0, self.succ_completion.call_count)

        def test_thread_op_a_times_out(self):
            cfg.CONF.set_override('request_poll_timeout', 1, group='octavia')
            self.driver.req.get.side_effect = [
                {'provisioning_status': 'PENDING_CREATE'}
            ]
            driver.thread_op(self.driver.load_balancer, self.lb)
            self.fail_completion.assert_called_once_with(self.context, self.lb)
            self.assertEqual(0, self.succ_completion.call_count)

        def test_thread_op_updates_vip_when_vip_delegated(self):
            cfg.CONF.set_override('allocates_vip', True, group='octavia')
            expected_vip = '10.1.1.1'
            self.driver.req.get.side_effect = [
                {'provisioning_status': 'PENDING_CREATE',
                 'vip': {'ip_address': ''}},
                {'provisioning_status': 'ACTIVE',
                 'vip': {'ip_address': expected_vip}}
            ]
            driver.thread_op(self.driver.load_balancer,
                             self.lb,
                             lb_create=True)
            self.succ_completion.assert_called_once_with(self.context, self.lb,
                                                         delete=False,
                                                         lb_create=True)
            self.assertEqual(expected_vip, self.lb.vip_address)


class TestOctaviaRequest(test_db_loadbalancerv2.LbaasPluginDbTestCase):

    def setUp(self):
        super(TestOctaviaRequest, self).setUp()
        self.OctaviaRequestObj = driver.OctaviaRequest("TEST_URL",
                                                       "TEST_SESSION")

    @mock.patch('requests.request')
    def test_request(self, mock_requests):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.content = "TEST"
        mock_response.headers = "TEST headers"
        mock_response.json.return_value = "TEST json"

        mock_requests.return_value = mock_response

        fake_header = {'X-Auth-Token': "TEST_TOKEN"}
        result = self.OctaviaRequestObj.request("TEST_METHOD", "TEST_URL",
                                                headers=fake_header)

        self.assertEqual("TEST json", result)

    @mock.patch('requests.request')
    def _test_request_Octavia(self, status_code, exception, mock_requests):
        mock_response = mock.MagicMock()
        mock_response.status_code = status_code
        mock_response.content = '{"faultstring": "FAILED"}'
        mock_response.headers = "TEST headers"

        mock_requests.return_value = mock_response

        fake_header = {'X-Auth-Token': "TEST_TOKEN"}
        self.assertRaisesRegex(exception,
                               "FAILED",
                               self.OctaviaRequestObj.request,
                               "TEST_METHOD", "TEST_URL", headers=fake_header)

        mock_response.json.assert_not_called()

    def test_request_Octavia_400(self):
        self._test_request_Octavia(400, exceptions.BadRequestException)

    def test_request_Octavia_401(self):
        self._test_request_Octavia(401, exceptions.NotAuthorizedException)

    def test_request_Octavia_403(self):
        self._test_request_Octavia(403, exceptions.NotAuthorizedException)

    def test_request_Octavia_404(self):
        self._test_request_Octavia(404, exceptions.NotFoundException)

    def test_request_Octavia_409(self):
        self._test_request_Octavia(409, exceptions.ConflictException)

    def test_request_Octavia_500(self):
        self._test_request_Octavia(500, exceptions.UnknownException)

    def test_request_Octavia_503(self):
        self._test_request_Octavia(503, exceptions.ServiceUnavailableException)



#  Copyright 2015, Shane McGough, KEMPtechnologies
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from kemptech_openstack_lbaas import driver as kemptech
from oslo_config import cfg

from neutron_lbaas.drivers import driver_base
from neutron_lbaas.drivers.kemptechnologies import config

cfg.CONF.register_opts(config.KEMP_OPTS, 'kemptechnologies')
CONF = cfg.CONF.kemptechnologies


class LoadBalancerManager(driver_base.BaseLoadBalancerManager):

    def create(self, context, lb):
        self.driver.kemptech.load_balancer.create(context, lb)

    def update(self, context, old_lb, lb):
        self.driver.kemptech.load_balancer.update(context, old_lb, lb)

    def delete(self, context, lb):
        self.driver.kemptech.load_balancer.delete(context, lb)

    def refresh(self, context, lb):
        self.driver.kemptech.load_balancer.refresh(context, lb)

    def stats(self, context, lb):
        return self.driver.kemptech.load_balancer.stats(context, lb)


class ListenerManager(driver_base.BaseListenerManager):

    def create(self, context, listener):
        self.driver.kemptech.listener.create(context, listener)

    def update(self, context, old_listener, listener):
        self.driver.kemptech.listener.update(context, old_listener, listener)

    def delete(self, context, listener):
        self.driver.kemptech.listener.delete(context, listener)


class PoolManager(driver_base.BasePoolManager):

    def create(self, context, pool):
        self.driver.kemptech.pool.create(context, pool)

    def update(self, context, old_pool, pool):
        self.driver.kemptech.pool.update(context, old_pool, pool)

    def delete(self, context, pool):
        self.driver.kemptech.pool.delete(context, pool)


class MemberManager(driver_base.BaseMemberManager):

    def create(self, context, member):
        self.driver.kemptech.member.create(context, member)

    def update(self, context, old_member, member):
        self.driver.kemptech.member.update(context, old_member, member)

    def delete(self, context, member):
        self.driver.kemptech.member.delete(context, member)


class HealthMonitorManager(driver_base.BaseHealthMonitorManager):

    def create(self, context, health_monitor):
        self.driver.kemptech.health_monitor.create(context, health_monitor)

    def update(self, context, old_health_monitor, health_monitor):
        self.driver.kemptech.health_monitor.update(context,
                                                   old_health_monitor,
                                                   health_monitor)

    def delete(self, context, health_monitor):
        self.driver.kemptech.health_monitor.delete(context, health_monitor)


class KempLoadMasterDriver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(KempLoadMasterDriver, self).__init__(plugin)
        self.load_balancer = LoadBalancerManager(self)
        self.listener = ListenerManager(self)
        self.pool = PoolManager(self)
        self.member = MemberManager(self)
        self.health_monitor = HealthMonitorManager(self)
        self.kemptech = kemptech.KempLoadMasterDriver(self, CONF)
